"""Transfer function fitting class
"""
import scipy.optimize

from .curvefit import CurveFit
import kontrol.curvefit.cost
import kontrol.curvefit.error_func


class TransferFunctionFit(CurveFit):
    """Transfer function fitting class

    This class is basically a ``CurveFit`` class
    with default cost functions and optimizer
    that is designed for fitting a transfer
    function.
    By default, the error function is
    ``kontrol.curvefit.error_func.tf_error()``.
    The default optimizer is ``scipy.optimize.minimize(
    ...,method="Nelder-Mead",...)``, with
    ``options = {"adaptive": True, "maxiter": N*1000}``,
    where ``N`` is the number of variables.
    All of these can be overridden if specified.

    Parameters
    ----------
    xdata : ``array`` or ``None``, optional
        Independent variable data.
        Defaults to ``None``.
    ydata : ``array`` or ``None``, optional
        Transfer function frequency response in complex numbers.
        Defaults to ``None``.
    model : ``func(x: ``array``, args: ``array``, **kwargs)`` ->\
            ``array``, or ``None``, optional
        The model used to fit the data.
        ``args`` in model is an ``array`` of parameters that
        define the model.
        Defaults to ``None``
    model_kwargs : ``dict`` or ``None``, optional
        Keyword arguments passed to the model.
        Defaults to ``None``.
    cost : ``kontrol.curvefit.Cost`` or ``func(args, model, xdata, ydata)``\
            -> ``array``, optional
        Cost function.
        The cost function to be used to fit the data.
        First argument is a list of parameters that will be passed to
        the model.
        This must be pickleable if multiprocessing is to be used.
        Defaults to ``None``.
    weight : ``array`` or ``None``, optional
        Weighting function.
        Defaults ``None``.
    error_func_kwargs : ``dict`` or ``None``, optional
        Keyword arguments the will be passed to ``error_func``,
        which is passed to the construct the cost function.
        Defaults to ``None``.
    optimizer : ``func(func, **kwargs)`` -> ``OptimizeResult``, or ``None``,\
                optional
        The optimization algorithm use for minimizing the cost function.
    optimizer_kwargs : ``dict`` or ``None``, optional
        Keyword arguments passed to the optimizer function.
        Defaults to ``None``.
    options : ``dict`` or ``None``, optional
        The option arguments passed to the ``optimizer``
    x0 : ``array``, optional
        Inital guess.
        Defaults to None.
    """
    def __init__(self, xdata=None, ydata=None,
                 model=None, model_kwargs=None,
                 cost=None, weight=None, error_func_kwargs=None,
                 optimizer=None, optimizer_kwargs=None, options=None,
                 x0=None,):
        """Constructor

        Parameters
        ----------
        xdata : ``array`` or ``None``, optional
            Independent variable data.
            Defaults to ``None``.
        ydata : ``array`` or ``None``, optional
            Transfer function frequency response in complex numbers.
            Defaults to ``None``.
        model : ``func(x: ``array``, args: ``array``, **kwargs)`` ->\
                ``array``, or ``None``, optional
            The model used to fit the data.
            ``args`` in model is an ``array`` of parameters that
            define the model.
            Defaults to ``None``
        model_kwargs : ``dict`` or ``None``, optional
            Keyword arguments passed to the model.
            Defaults to ``None``.
        cost : ``kontrol.curvefit.Cost`` or ``func(args, model, xdata, ydata)``\
                -> ``array``, optional
            Cost function.
            The cost function to be used to fit the data.
            First argument is a list of parameters that will be passed to
            the model.
            This must be pickleable if multiprocessing is to be used.
            Defaults to ``None``.
        weight : ``array`` or ``None``, optional
            Weighting function.
            Defaults ``None``.
        error_func_kwargs : ``dict`` or ``None``, optional
            Keyword arguments the will be passed to ``error_func``,
            which is passed to the construct the cost function.
            Defaults to ``None``.
        optimizer : ``func(func, **kwargs)`` -> ``OptimizeResult``, or ``None``,\
                    optional
            The optimization algorithm use for minimizing the cost function.
        optimizer_kwargs : ``dict`` or ``None``, optional
            Keyword arguments passed to the optimizer function.
            Defaults to ``None``.
        options : ``dict`` or ``None``, optional
            The option arguments passed to the ``optimizer``
        x0 : ``array``, optional
            Inital guess.
            Defaults to None.
        """
        if cost is None:
            error_func = kontrol.curvefit.error_func.tf_error
            default_error_func_kwargs = {"weight": weight}
            if error_func_kwargs is None:
                error_func_kwargs = default_error_func_kwargs
            else:
                error_func_kwargs = dict(
                    default_error_func_kwargs, **error_func_kwargs)
            cost = kontrol.curvefit.cost.Cost(
                error_func=error_func, error_func_kwargs=error_func_kwargs)

        if optimizer is None:
            optimizer = scipy.optimize.minimize

        if x0 is None:
            maxiter = None
        else:
            maxiter = len(x0)*1000
        default_options = {"adaptive": True, "maxiter": maxiter}    
        if options is None:
            options = default_options
        else:
            options = dict(default_options, **options)

        default_optimizer_kwargs = {
            "x0": x0, "method": "Nelder-Mead", "options": options}
        if optimizer_kwargs is None:
            optimizer_kwargs = default_optimizer_kwargs
        else:
            if "options" in optimizer_kwargs:
                options = dict(default_options, **optimizer_kwargs["options"])
                optimizer_kwargs["options"] = options
            optimizer_kwargs = dict(
                default_optimizer_kwargs, **optimizer_kwargs)

        super().__init__(
            xdata, ydata, model, model_kwargs,
            cost, optimizer, optimizer_kwargs)

        self._weight = None
        self._options = None
        self._x0 = None
        self.weight = weight
        self.options = options
        self.x0 = x0

    @property
    def weight(self):
        """Weighting function"""
        return self._weight

    @weight.setter
    def weight(self, _weight):
        """weight.setter"""
        self._weight = _weight
        error_func_kwargs = {"weight": self.weight}
        self.cost.error_func_kwargs = dict(
            self.cost.error_func_kwargs, **error_func_kwargs)

    @property
    def options(self):
        """Option arguments passed to optimizer"""
        return self._options
    
    @options.setter
    def options(self, _options):
        """options.setter"""
        if _options is None:
            self._options = {}
        else:
            self._options = _options
        self._update_optimizer_kwargs()
        
    @property
    def x0(self):
        """Initial guess"""
        return self._x0
    
    @x0.setter
    def x0(self, _x0):
        """x0.setter"""
        self._x0 = _x0
        self._update_optimizer_kwargs()

    def _get_maxiter(self):
        """Get `maxiter` option from the initial guess `x0`"""
        # Added this method so we can change the multiplier here.
        if self.x0 is None:
            return None
        else:
            return len(self.x0)*1000

    def _update_optimizer_kwargs(self):
        """Set optimizer_kwargs"""
        if self.x0 is not None:
            update_optimizer_kwargs = {"x0": self.x0}
            self.optimizer_kwargs = dict(
                self.optimizer_kwargs, **update_optimizer_kwargs)
        self._update_options()
    
    def _update_options(self):
        """Update optimizer options"""
        maxiter = self._get_maxiter()
        update_options = {"maxiter": maxiter}    
        if "options" not in self.optimizer_kwargs:
            options = update_options
        else: 
            options = dict(self.optimizer_kwargs["options"], **update_options)
        options = dict(options, **self.options)  # self.options have priority
        self.optimizer_kwargs["options"] = options




"""Filter synthesis functions.
"""
import control

import kontrol.core.controlutils


def generalized_plant(noise1, noise2, weight1, weight2):
    """Return the generalized plant of a 2 complementary filter system
    
    Parameters
    ----------
    noise1 : TransferFunction
        Sensor noise 1 transfer function model.
        A transfer function that has magnitude response matching the
        amplitude spectral density of noise 1.
    noise2 : TransferFunction
        Sensor noise 2 transfer function model.
        A transfer function that has magnitude response matching the
        amplitude spectral density of noise 2.
    weight1 : TransferFunction
        Weighting function 1.
        Frequency dependent specification for noise 1.
    weight2 : TransferFunction
        Weighting function 2.
        Frequency dependent specification for noise 2.

    Returns
    -------
    control.xferfcn.TransferFunction
        The plant.
    """
    zero = control.tf([0], [1])
    one = control.tf([1], [1])
    if weight1 is None:
        weight1 = one
    if weight2 is None:
        weight2 = one

    p = [[zero, weight2*noise2, one],
         [weight1*noise1, -weight2*noise2, zero]]
    # Tall matrix version here, not used.
#     p = [[noise1, -noise1],
#          [tf([0],[1]), noise2],
#          [tf([1],[1]), tf([0],[1])]]
    p = kontrol.core.controlutils.tfmatrix2tf(p)
    return p
    

def h2complementary(noise1, noise2, weight1=None, weight2=None):
    """H2 optimal complementary filter synthesis

    Parameters
    ----------
    noise1 : TransferFunction
        Sensor noise 1 transfer function model.
        A transfer function that has magnitude response matching the
        amplitude spectral density of noise 1.
    noise2 : TransferFunction
        Sensor noise 2 transfer function model.
        A transfer function that has magnitude response matching the
        amplitude spectral density of noise 2.
    weight1 : TransferFunction, optional
        Weighting function 1.
        Frequency dependent specification for noise 1.
        Defaults None.
    weight2 : TransferFunction, optional
        Weighting function 2.
        Frequency dependent specification for noise 2.

    Returns
    -------
    filter1 : TransferFunction
        The complementary filter filtering noise1.
    filter2 : TransferFunction
        The complementary filter filtering noise2.

    Notes
    -----
    This function ultilizes control.robust.h2syn() which depends on the slycot
    module. If you are using under a conda virtual environment, the slycot
    module can be installed easily from conda-forge. Using pip to install
    slycot is a bit more involved (I have yet to suceed installing slycot in
    my Windows machine). Please refer to the python-control package for further
    instructions.

    It is possible that h2syn yields no solution for some tricky noise profiles
    . Try adjusting the noise profiles at some irrelevant frequencies.

    Thomas Dehaeze [1]_ had the idea first so credits goes to him. (Properly
    cite when the paper is published.)

    References
    ----------
    .. [1]
        Dehaeze, T.
        https://tdehaeze.github.io/dehaeze20_optim_robus_compl_filte/matlab/\
index.html
    .. [2]
        T. T. L. Tsang, T. G. F. Li, T. Dehaeze, C. Collette.
        Optimal Sensor Fusion Method for Active Vibration Isolation Systems in
        Ground-Based Gravitational-Wave Detectors.
        https://arxiv.org/pdf/2111.14355.pdf
    """
    p = generalized_plant(noise1, noise2, weight1, weight2)
    p = control.ss(p)
    filter1 = control.h2syn(p, 1, 1)
    filter1 = control.tf(filter1)
    filter2 = 1 - filter1
    return filter1, filter2


def hinfcomplementary(noise1, noise2, weight1=None, weight2=None):
    """H-infinity optimal complementary filter synthesis

    Parameters
    ----------
    noise1 : TransferFunction
        Sensor noise 1 transfer function model.
        A transfer function that has magnitude response matching the
        amplitude spectral density of noise 1.
    noise2 : TransferFunction
        Sensor noise 2 transfer function model.
        A transfer function that has magnitude response matching the
        amplitude spectral density of noise 2.
    weight1 : TransferFunction, optional
        Weighting function 1.
        Frequency dependent specification for noise 1.
        Defaults None.
    weight2 : TransferFunction, optional
        Weighting function 2.
        Frequency dependent specification for noise 2.

    Returns
    -------
    filter1 : TransferFunction
        The complementary filter filtering noise1.
    filter2 : TransferFunction
        The complementary filter filtering noise2.

    Notes
    -----
    This function ultilizes control.robust.hinfsyn() which depends on the
    slycot module. If you are using under a conda virtual environment,
    the slycot module can be installed easily from conda-forge.
    Using pip to install slycot is a bit more involved
    (I have yet to suceed installing slycot in
    my Windows machine).
    Please refer to the python-control package for further instructions.

    Thomas Dehaeze [1]_ had the idea first so credits goes to him. (Properly
    cite when the paper is published.)

    References
    ----------
    .. [1]
        Dehaeze, T.
        https://tdehaeze.github.io/dehaeze20_optim_robus_compl_filte/matlab/\
index.html
    .. [2]
        T. T. L. Tsang, T. G. F. Li, T. Dehaeze, C. Collette.
        Optimal Sensor Fusion Method for Active Vibration Isolation Systems in
        Ground-Based Gravitational-Wave Detectors.
        https://arxiv.org/pdf/2111.14355.pdf
    """
    p = generalized_plant(noise1, noise2, weight1, weight2)
    p = control.ss(p)
    filter1, _, gamma, _ = control.hinfsyn(p, 1, 1)
    # print(gamma)
    filter1 = control.tf(filter1)
    filter2 = 1 - filter1
    return filter1, filter2



"""Ad Hoc class for Easy channel access (EZCA).
"""
import ezca
import numpy as np


class Ezca(ezca.Ezca):
    """Ezca class for access channel variables
    """
    def __init__(self, prefix, timeout=0.1):
        """Constructor

        Parameters
        ----------
        prefix : str
            Channel prefix. E.g. "K1:VIS-BS".
            "K1:" is implied in KAGRA DGS so "VIS-BS" is equivalent to
            "K1:VIS-BS".
        timeout : int, optional
            Timeout in seconds. Defaults to 0.1.
        """
        super().__init__(prefix, timeout=timeout)

    def get_matrix(self, matrix, row_slicers=None, column_slicers=None):
        """Returns a matrix from the EPICS record using Ezca.

        Parameters
        ----------
        matrix : str
            The matrix to be accessed.
        row_slicers : int, (int, int) or None, optional
            Row slicers, the rows to be accessed. Counting from 1.
            If an int is specified, then the number of rows to be accessed
            will be this number, starting from the [1, 1] element.
            If a submatrix is to be accessed, then specify as a tuple.
            Defaults None. If None, sets to full matrix.
        column_slicers : int, (int, int) or None, optional
            Column slicers, the columns to be accessed. Counting from 1.
            If an int is specified, then the number of columns to be accessed
            will be this number, starting from the [1, 1] element.
            If a submatrix is to be accessed, then specify as a tuple.
            Defaults None. If None, sets to full matrix.

        Returns
        -------
        numpy.array
            The accessed matrix.
        """
        # Read the matrix until error to get the number of rows and columns.
        nrow, ncol = self._get_row_column(matrix)
        if isinstance(row_slicers, int):
            row_slicers = (1, row_slicers)
        elif row_slicers is None:
            row_slicers = (1, nrow)
        if isinstance(column_slicers, int):
            column_slicers = (1, column_slicers)
        elif column_slicers is None:
            column_slicers = (1, ncol)

        ezca_matrix = np.zeros((nrow, ncol))  # Initialize matrix placeholder
        for i in range(len(ezca_matrix)):
            for j in range(len(ezca_matrix)):
                ezca_matrix[i, j] = self.read(
                    "{}_{}_{}".format(matrix, i+1, j+1))

        return ezca_matrix[row_slicers[0]-1:row_slicers[1],
                           column_slicers[0]-1:column_slicers[1]]

    def put_matrix(self, numpy_array, matrix,
                   row_slicers=None, column_slicers=None):
        """Put a matrix into the EPICS record using Ezca.

        Parameters
        ----------
        numpy_array : array
            The matrix values to be put.
        matrix : str
            The matrix in EPICS record.
        row_slicers : (int, int) or None, optional
            Row slicers, the rows to be accessed. Counting from 1.
            The size determined by the tuple must match the size of
            `numpy_array`
            Defaults None. If None, sets to full matrix.
        column_slicers : (int, int) or None, optional
            Column slicers, the columns to be accessed. Counting from 1.
            The size determined by the tuple must match the size of
            `numpy_array`
            Defaults None. If None, sets to full matrix.
        """
        # Read the matrix until error to get the number of rows and columns.
        nrow, ncol = self._get_row_column(matrix)
        if row_slicers is None:
            row_slicers = (1, nrow)
        if column_slicers is None:
            column_slicers = (1, ncol)

        if row_slicers[1] - row_slicers[0] + 1 != len(numpy_array):
            raise ValueError("Slicers must match the size of numpy_array.")
        elif column_slicers[1] - column_slicers[0] + 1 != len(numpy_array[0]):
            raise ValueError("Slicers must match the size of numpy_array.")

        epics_matrix = self.get_matrix(matrix)
        epics_matrix[row_slicers[0]-1:row_slicers[1],
                     column_slicers[0]-1: column_slicers[1]] = numpy_array

        self._put_matrix(numpy_array=epics_matrix, matrix=matrix)

    def _put_matrix(self, numpy_array, matrix):
        """Put a matrix into the EPICS record using Ezca, (low level function).

        Parameters
        ----------
        numpy_array : array
            The matrix values to be put.
            Matrix size must match that in the EPICS record.
        matrix : str
            The matrix in EPICS record.
        """
        nrow, ncol = self._get_row_column(matrix)
        if len(numpy_array) != nrow or len(numpy_array[0]) != ncol:
            raise ValueError("Size of numpy_array not equal to that in the"
                             " EPICS record.")

        for i in range(len(numpy_array)):
            for j in range(len(numpy_array[i])):
                self.write("{}_{}_{}".format(matrix, i+1, j+1),
                           numpy_array[i, j])


    def _get_row_column(self, matrix):
        """Get the number of rows and columns of a matrix.

        Parameters
        ==========
        matrix : str
            The matrix to be accessed.

        Returns
        =======
        int, int
            The number of rows and columns of the matrix.
        """
        # Try increasing the number of row and columns numbers.
        i = 1
        j = 1
        while 1:
            try:
                self.read("{}_{}_1".format(matrix, i))
                i += 1
            except ezca.errors.EzcaConnectError:
                if i > 0:
                    i -= 1
                break
        while 1:
            try:
                self.read("{}_1_{}".format(matrix, j))
                j += 1
            except ezca.errors.EzcaConnectError:
                if j > 0:
                    j -= 1
                break
        return i, j



"""Spectrum fitting class
"""
import numpy as np
import scipy.optimize

from .curvefit import CurveFit
import kontrol.curvefit.cost
import kontrol.curvefit.error_func
import kontrol.curvefit.model


class SpectrumZPKFit(CurveFit):
    """Spectrum ZPK fitting class

    This class is basically a ``CurveFit`` class
    with default cost functions and optimizer
    that is designed for fitting a spectrum with
    a simple ZPK model.
    By default, the error function is
    ``kontrol.curvefit.error_func.spectrum_error()``.
    The default optimizer is ``scipy.optimize.differential_evolution(), with
    ``options = {"bounds": bounds, "workers": -1, "updating"="deferred"}``,
    where ``bounds`` is [(min(xdata), max(xdata)] * (nzeros + npoles),
    and appended with [(min(ydata), max(ydata)].
    The bounds are np.log10() if log_args is true in the model.
    All of these can be overridden if specified.

    Parameters
    ----------
    xdata : ``array`` or ``None``, optional
        Independent variable data.
        Defaults to ``None``.
    ydata : ``array`` or ``None``, optional
        Transfer function frequency response in complex numbers.
        Defaults to ``None``.
    model : ``func(x: ``array``, args: ``array``, **kwargs)`` ->\
            ``array``, or ``None``, optional
        The model used to fit the data.
        ``args`` in model is an ``array`` of parameters that
        define the model.
        Defaults to ``None``
    model_kwargs : ``dict`` or ``None``, optional
        Keyword arguments passed to the model.
        Defaults to ``None``.
    cost : ``kontrol.curvefit.Cost`` or ``func(args, model, xdata, ydata)``\
            -> ``array``, optional
        Cost function.
        The cost function to be used to fit the data.
        First argument is a list of parameters that will be passed to
        the model.
        This must be pickleable if multiprocessing is to be used.
        Defaults to ``None``.
    weight : ``array`` or ``None``, optional
        Weighting function.
        Defaults ``None``.
    error_func_kwargs : ``dict`` or ``None``, optional
        Keyword arguments the will be passed to ``error_func``,
        which is passed to the construct the cost function.
        Defaults to ``None``.
    optimizer : ``func(func, **kwargs)`` -> ``OptimizeResult``, or ``None``,\
                optional
        The optimization algorithm use for minimizing the cost function.
    optimizer_kwargs : ``dict`` or ``None``, optional
        Keyword arguments passed to the optimizer function.
        Defaults to ``None``.
    seed : int, optional
        The random seed.
        Defaults 123.
    """
    def __init__(self, xdata=None, ydata=None,
                 model=None, model_kwargs=None, npole=None, nzero=None,
                 cost=None, weight=None, error_func_kwargs=None,
                 optimizer=None, optimizer_kwargs=None, seed=123):
        """Constructor

        Parameters
        ----------
        xdata : ``array`` or ``None``, optional
            Independent variable data.
            Defaults to ``None``.
        ydata : ``array`` or ``None``, optional
            Transfer function frequency response in complex numbers.
            Defaults to ``None``.
        model : ``func(x: ``array``, args: ``array``, **kwargs)`` ->\
                ``array``, or ``None``, optional
            The model used to fit the data.
            ``args`` in model is an ``array`` of parameters that
            define the model.
            Defaults to ``None``
        model_kwargs : ``dict`` or ``None``, optional
            Keyword arguments passed to the model.
            Defaults to ``None``.
        cost : ``kontrol.curvefit.Cost`` or ``func(args, model, xdata, ydata)``\
                -> ``array``, optional
            Cost function.
            The cost function to be used to fit the data.
            First argument is a list of parameters that will be passed to
            the model.
            This must be pickleable if multiprocessing is to be used.
            Defaults to ``None``.
        weight : ``array`` or ``None``, optional
            Weighting function.
            Defaults ``None``.
        error_func_kwargs : ``dict`` or ``None``, optional
            Keyword arguments the will be passed to ``error_func``,
            which is passed to the construct the cost function.
            Defaults to ``None``.
        optimizer : ``func(func, **kwargs)`` -> ``OptimizeResult``, or ``None``,\
                    optional
            The optimization algorithm use for minimizing the cost function.
        optimizer_kwargs : ``dict`` or ``None``, optional
            Keyword arguments passed to the optimizer function.
            Defaults to ``None``.
        seed : int, optional
            The random seed.
            Defaults 123.
        """
        if cost is None:
            error_func = kontrol.curvefit.error_func.spectrum_error
            default_error_func_kwargs = {"weight": weight}
            if error_func_kwargs is None:
                error_func_kwargs = default_error_func_kwargs
            else:
                error_func_kwargs = dict(
                    default_error_func_kwargs, **error_func_kwargs)
            cost = kontrol.curvefit.cost.Cost(
                error_func=error_func, error_func_kwargs=error_func_kwargs)

        if optimizer is None:
            optimizer = scipy.optimize.differential_evolution

        if (model is not None and xdata is not None and ydata is not None) :
            n_zero_pole_bounds = model.nzero + model.npole
            if model.log_args: 
                bound = (np.log10(min(xdata)), np.log10(max(xdata)))
                bounds = [bound] * n_zero_pole_bounds
                bounds.append((np.log10(min(ydata)), np.log10(max(ydata))))
            else:
                bound = (min(xdata), max(xdata))
                bounds = [bound] * n_zero_pole_bounds
                bounds.append((min(ydata), max(ydata)))
        else:
            bounds = None

        default_optimizer_kwargs = {
            "workers": -1, "updating": "deferred", "bounds": bounds}    
        if optimizer_kwargs is None:
            optimizer_kwargs = default_optimizer_kwargs
        else:
            optimizer_kwargs = dict(
                default_optimizer_kwargs, **optimizer_kwargs)

        super().__init__(
            xdata, ydata, model, model_kwargs,
            cost, optimizer, optimizer_kwargs)
        self._weight = None
        self.weight = weight
        self._model = None
        self.model = model
        self._seed = None
        self.seed = seed

    @property
    def seed(self):
        """Random seed"""
        return self._seed

    @seed.setter
    def seed(self, _seed):
        """seed setter"""
        self._seed = _seed

    @property
    def weight(self):
        """Weighting function"""
        return self._weight

    @weight.setter
    def weight(self, _weight):
        """weight.setter"""
        self._weight = _weight
        error_func_kwargs = {"weight": self.weight}
        self.cost.error_func_kwargs = dict(
            self.cost.error_func_kwargs, **error_func_kwargs)

    @property
    def xdata(self):
        """xdata"""
        return self._xdata
    
    @xdata.setter
    def xdata(self, _xdata):
        """xdata setter"""
        self._xdata = _xdata
        self._update_optimizer_kwargs()

    @property
    def ydata(self):
        """ydata"""
        return self._ydata
    
    @ydata.setter
    def ydata(self, _ydata):
        """ydata setter"""
        self._ydata = _ydata
        self._update_optimizer_kwargs()

    @property
    def model(self):
        """The ZPK model used to fit the spectrum"""
        return self._model
    
    @model.setter
    def model(self, _model):
        """model setter"""
        self._model = _model
        self._update_optimizer_kwargs()

    def _get_bounds(self):
        """Get bounds for scipy.optimize.differential_evolution"""
        if (self.model is not None and self.xdata is not None
            and self.ydata is not None):
            # 
            n_zero_pole_bounds = self.model.nzero + self.model.npole
            if self.model.log_args: 
                bound = (np.log10(min(self.xdata)), np.log10(max(self.xdata)))
                bounds = [bound] * n_zero_pole_bounds
                bounds.append(
                    (np.log10(min(self.ydata)), np.log10(max(self.ydata))))
            else:
                bound = (min(self.xdata), max(self.xdata))
                bounds = [bound] * n_zero_pole_bounds
                bounds.append((min(self.ydata), max(self.ydata)))
        else:
            bounds = None
        return bounds

    def _update_optimizer_kwargs(self):
        """Set optimizer_kwargs"""
        bounds = self._get_bounds()
        if bounds is not None:
            update_optimizer_kwargs = {"bounds": bounds}
            self.optimizer_kwargs = dict(
                self.optimizer_kwargs, **update_optimizer_kwargs)

    def prefit(self):
        """Something to do before fitting"""
        np.random.seed(self.seed)


class SpectrumTFFit(CurveFit):
    """Spectrum Transfer function fitting class

    This class is basically a ``CurveFit`` class
    with default cost functions and optimizer
    that is designed for fitting a spectrum with a transfer function
    By default, the error function is
    ``kontrol.curvefit.error_func.spectrum_error()``.
    The default optimizer is ``scipy.optimize.minimize(
    ...,method="Nelder-Mead",...)``, with
    ``options = {"adaptive": True, "maxiter": N*1000}``,
    where ``N`` is the number of variables.
    All of these can be overridden if specified.

    Parameters
    ----------
    xdata : ``array`` or ``None``, optional
        Independent variable data.
        Defaults to ``None``.
    ydata : ``array`` or ``None``, optional
        Transfer function frequency response in complex numbers.
        Defaults to ``None``.
    model : ``func(x: ``array``, args: ``array``, **kwargs)`` ->\
            ``array``, or ``None``, optional
        The model used to fit the data.
        ``args`` in model is an ``array`` of parameters that
        define the model.
        Defaults to ``None``
    model_kwargs : ``dict`` or ``None``, optional
        Keyword arguments passed to the model.
        Defaults to ``None``.
    cost : ``kontrol.curvefit.Cost`` or ``func(args, model, xdata, ydata)``\
            -> ``array``, optional
        Cost function.
        The cost function to be used to fit the data.
        First argument is a list of parameters that will be passed to
        the model.
        This must be pickleable if multiprocessing is to be used.
        Defaults to ``None``.
    weight : ``array`` or ``None``, optional
        Weighting function.
        Defaults ``None``.
    error_func_kwargs : ``dict`` or ``None``, optional
        Keyword arguments the will be passed to ``error_func``,
        which is passed to the construct the cost function.
        Defaults to ``None``.
    optimizer : ``func(func, **kwargs)`` -> ``OptimizeResult``, or ``None``,\
                optional
        The optimization algorithm use for minimizing the cost function.
    optimizer_kwargs : ``dict`` or ``None``, optional
        Keyword arguments passed to the optimizer function.
        Defaults to ``None``.
    options : ``dict`` or ``None``, optional
        The option arguments passed to the ``optimizer``
    num0 : ``array``, optional
        Inital guess of the numerator.
        Defaults to None.
    den0 : ``array``, optional
        Inital guess of the denominator.
        Defaults to None.
    """
    def __init__(self, xdata=None, ydata=None,
                 model=None, model_kwargs=None,
                 cost=None, weight=None, error_func_kwargs=None,
                 optimizer=None, optimizer_kwargs=None, options=None,
                 num0=None, den0=None):
        """Constructor

        Parameters
        ----------
        xdata : ``array`` or ``None``, optional
            Independent variable data.
            Defaults to ``None``.
        ydata : ``array`` or ``None``, optional
            Transfer function frequency response in complex numbers.
            Defaults to ``None``.
        model : ``func(x: ``array``, args: ``array``, **kwargs)`` ->\
                ``array``, or ``None``, optional
            The model used to fit the data.
            ``args`` in model is an ``array`` of parameters that
            define the model.
            Defaults to ``None``
        model_kwargs : ``dict`` or ``None``, optional
            Keyword arguments passed to the model.
            Defaults to ``None``.
        cost : ``kontrol.curvefit.Cost`` or ``func(args, model, xdata, ydata)``\
                -> ``array``, optional
            Cost function.
            The cost function to be used to fit the data.
            First argument is a list of parameters that will be passed to
            the model.
            This must be pickleable if multiprocessing is to be used.
            Defaults to ``None``.
        weight : ``array`` or ``None``, optional
            Weighting function.
            Defaults ``None``.
        error_func_kwargs : ``dict`` or ``None``, optional
            Keyword arguments the will be passed to ``error_func``,
            which is passed to the construct the cost function.
            Defaults to ``None``.
        optimizer : ``func(func, **kwargs)`` -> ``OptimizeResult``, or ``None``,\
                    optional
            The optimization algorithm use for minimizing the cost function.
        optimizer_kwargs : ``dict`` or ``None``, optional
            Keyword arguments passed to the optimizer function.
            Defaults to ``None``.
        options : ``dict`` or ``None``, optional
            The option arguments passed to the ``optimizer``
        num0 : ``array``, optional
            Inital guess of the numerator.
            Defaults to None.
        den0 : ``array``, optional
            Inital guess of the denominator.
            Defaults to None.
        """
        if cost is None:
            error_func = kontrol.curvefit.error_func.spectrum_error
            default_error_func_kwargs = {"weight": weight}
            if error_func_kwargs is None:
                error_func_kwargs = default_error_func_kwargs
            else:
                error_func_kwargs = dict(
                    default_error_func_kwargs, **error_func_kwargs)
            cost = kontrol.curvefit.cost.Cost(
                error_func=error_func, error_func_kwargs=error_func_kwargs)

        if optimizer is None:
            optimizer = scipy.optimize.minimize

        if num0 is None or den0 is None:
            x0 = None
            maxiter = None
        else:
            x0 = np.append(num0, den0)
            maxiter = len(x0)*1000
             
        default_options = {"adaptive": True, "maxiter": maxiter}    
        if options is None:
            options = default_options
        else:
            options = dict(default_options, **options)

        default_optimizer_kwargs = {
            "x0": x0, "method": "Nelder-Mead", "options": options}
        if optimizer_kwargs is None:
            optimizer_kwargs = default_optimizer_kwargs
        else:
            if "options" in optimizer_kwargs:
                options = dict(default_options, **optimizer_kwargs["options"])
                optimizer_kwargs["options"] = options
            optimizer_kwargs = dict(
                default_optimizer_kwargs, **optimizer_kwargs)

        super().__init__(
            xdata, ydata, model, model_kwargs,
            cost, optimizer, optimizer_kwargs)

        self._weight = None
        self._options = None
        self._x0 = None
        self._num0 = None
        self._den0 = None
        self.weight = weight
        self.options = options
        self.x0 = x0
        self.num0 = num0
        self.den0 = den0

    @property
    def weight(self):
        """Weighting function"""
        return self._weight

    @weight.setter
    def weight(self, _weight):
        """weight.setter"""
        self._weight = _weight
        error_func_kwargs = {"weight": self.weight}
        self.cost.error_func_kwargs = dict(
            self.cost.error_func_kwargs, **error_func_kwargs)

    @property
    def options(self):
        """Option arguments passed to optimizer"""
        return self._options
    
    @options.setter
    def options(self, _options):
        """options.setter"""
        if _options is None:
            self._options = {}
        else:
            self._options = _options
        self._update_optimizer_kwargs()

    @property
    def num0(self):
        """Initial numerator guess"""
        return self._num0

    @num0.setter
    def num0(self, _num0):
        """num0 setter"""
        self._num0 = _num0
        if self.num0 is not None and self.den0 is not None:
            self.x0 = np.append(self.num0, self.den0)

    @property
    def den0(self):
        """Initial denominator guess"""
        return self._den0

    @den0.setter
    def den0(self, _den0):
        """den0 setter"""
        self._den0 = _den0
        if self.num0 is not None and self.den0 is not None:
            self.x0 = np.append(self.num0, self.den0)
        
    @property
    def x0(self):
        """Initial guess"""
        return self._x0
    
    @x0.setter
    def x0(self, _x0):
        """x0.setter"""
        self._x0 = _x0
        self._update_optimizer_kwargs()

    def _get_maxiter(self):
        """Get `maxiter` option from the initial guess `x0`"""
        # Added this method so we can change the multiplier here.
        if self.x0 is None:
            return None
        else:
            return len(self.x0)*1000

    def _update_optimizer_kwargs(self):
        """Set optimizer_kwargs"""
        if self.x0 is not None:
            update_optimizer_kwargs = {"x0": self.x0}
            self.optimizer_kwargs = dict(
                self.optimizer_kwargs, **update_optimizer_kwargs)
        self._update_options()
    
    def _update_options(self):
        """Update optimizer options"""
        maxiter = self._get_maxiter()
        update_options = {"maxiter": maxiter}    
        if "options" not in self.optimizer_kwargs:
            options = update_options
        else: 
            options = dict(self.optimizer_kwargs["options"], **update_options)
        options = dict(options, **self.options)  # self.options have priority
        self.optimizer_kwargs["options"] = options


def spectrum_fit(f, spectrum, nzero, npole, log_args=True,
                 seed=123, minreal=True, clean=True, stabilize=True,
                 return_zpk_fit=False, return_tf_fit=False):
    """Fit a spectrum with a zpk model intermediately and then with a tf model.

    Parameters
    ----------
    f : array
        Frequency axis.
    spectrum : array
        The frequency spectrum.
    nzero : int
        Number of zeros in the transfer function.
    npole : int
        Number of poles in the transfer function.
    log_args : boolean, optional
        Fit the log10 of the arguments instead.
        Defaults True.
    seed : int, optional
        Random seed.
    minreal : boolean, optional
        Return the minreal version of the transfer function.
        Defaults True.
    clean : boolean, optional
        Clean the transfer function before returning.
        This uses kontrol.TransferFunction.clean() to remove outlier
        coefficients.
        Defaults True.
    stabilize : boolean, optional
        Negate the real parts of unstable zeros and poles.
        Defaults True.
    return_zpk_fit : boolean, optional
        Return the kontrol.curvefit.SpectrumZPKFit object.
        Defaults False.
    return_tf_fit : boolean, optional
        Return the kontrol.curvefit.SpectrumTFFit object.
        Defaults False.

    Returns
    -------
    tf : Kontrol.TransferFunction
        The fitted transfer function
    zpk_fit : Kontrol.curvefit.SpectrumZPKFit, optional
        The CurveFit object used to fit intermediately.
        Return only if ``return_zpk_fit`` is ``True``.
    tf_fit : Kontrol.curvefit.SpectrumTFFit, optional
        The CurveFit object used to fit the final transfer function
        Return only if ``return_tf_fit`` is ``True``.
    """
    # Fit a ZPK model.
    zpk_model = kontrol.curvefit.model.SimpleZPK(
        nzero=nzero, npole=npole, log_args=log_args)
    zpk_fit = SpectrumZPKFit(seed=seed)
    zpk_fit.xdata = f
    zpk_fit.ydata = spectrum
    zpk_fit.model = zpk_model
    zpk_fit.fit()

    # Extract the fitted zpk model's numerator and denominator
    tf_zpk_fit = zpk_fit.model.tf
    tf_zpk_fit = tf_zpk_fit.minreal()
    num0 = tf_zpk_fit.num[0][0]
    den0 = tf_zpk_fit.den[0][0]
    if log_args:
        num0 = np.log10(num0)
        den0 = np.log10(den0)

    # Fit a transfer function model using the fitted ZPK as an initial guess 
    tf_model = kontrol.curvefit.model.TransferFunctionModel(
        nzero=nzero, npole=npole, log_args=log_args)
    tf_fit = SpectrumTFFit()
    tf_fit.xdata = f
    tf_fit.ydata = spectrum
    tf_fit.model = tf_model
    tf_fit.num0 = num0
    tf_fit.den0 = den0
    tf_fit.fit()

    # Post processing
    tf = tf_fit.model.tf
    if stabilize:
        tf.stabilize()
    if clean:
        tf.clean()  # This actually does minreal already.
    if minreal:
        # minreal returns a control.TransferFunction object...
        tf = kontrol.TransferFunction(tf.minreal())
    
    # Return
    if return_zpk_fit and return_tf_fit:
        return tf, zpk_fit, tf_fit
    elif return_zpk_fit:
        return tf, zpk_fit
    elif return_tf_fit:
        return tf, tf_fit
    else:
        return tf

    



"""Frequency series class.
"""
import inspect

import numpy as np
import scipy.optimize

import kontrol.core.math
import kontrol.frequency_series.costs
import kontrol.frequency_series.conversion


class FrequencySeries:
    """ Frequency series class

    Attributes
    ----------
    f: array
        Frequency axis
    x: array
        Frequency series data.
    x_empirical: array or None
        Frequency series of the empirical model.
    model_empirical: func(f, a, b, c, ...) -> array or None
        The model function whose first argument is the frequency axis and
        the rest being an arbitrary number of model parameters to be fit.
    args_empirical_model: array or None
        The optimized arguments passed to the model_empirical.
    f_zpk: array or None
        The frequency axis used during ZPK fitting.
    x_zpk: array or None
        The frequency series data used during ZPK fitting.
        Note: this is a complex array.
    tf_zpk: control.xferfcn.TransferFunction or None
        The transfer function the best fit the data using ZPK model.
    args_zpk_model: array or None
        A 1-D list of zeros, poles, and gain.
        Zeros and poles are in unit of Hz.
    f_tf: array or None
        The frequency series data used during transfer function fitting.
    x_tf: array or None
        The frequency series data used during transfer function fitting.
        Note: this is a complex array.
    tf: control.xferfcn.TransferFunction or None
        The modeled transfer function.
    args_tf_model: array or None
        A 1-D list of numerator and denominator coefficients,
        from higher order to lower order.
    """
    def __init__(self, f, x):
        """Constructor

        Parameters
        ----------
        f: array
            Frequency axis
        x: array
            Frequency series data.
        """
        self.f = f
        self.x = x
        self.x_empirical = None
        self.model_empirical = None
        self.args_empirical_model = None
        self.f_zpk = None
        self.x_zpk = None
        self.tf_zpk = None
        self.args_zpk_model = None
        self._f_processed = None
        self._x_processed = None
        self.f_tf = None
        self.x_tf = None
        self.tf = None
        self.args_tf_model = None

    def fit_empirical(self, model, x0=None,
                      error_func=kontrol.core.math.log_mse,
                      error_func_kwargs={},
                      optimizer=scipy.optimize.minimize,
                      optimizer_kwargs={}):
        """Fit frequency series data with an empirical model. (Local optimize)

        Parameters
        ----------
        model: func(f, a, b, c, ...) -> array
            The model function whose first argument is the frequency axis and
            the rest being an arbitrary number of model parameters to be fit.
        x0: array, optional
            The initial parameters for fitting.
            Defaults None. If None, defaults to np.ones()
        error_func: func(x1: array, x2: array) -> float
            The function that evaluate the error between arrays x1 and x2.
            Defaults to kontrol.core.math.log_mse, which evaluates the
            logarithmic mean square error.
        error_func_kwargs: dict, optional
            Keyword arguments passed to the error function.
            Use this to specify weighting function {"weight": weight_func}.
            Defaults to {}.
        optimizer: func, optional
            The optimization function which has the signature
            func(func, args, bounds, **kw) -> scipy.optimize.OptimizeResult.
            Defaults to scipy.optimize.minimize.
        optimizer_kwargs: dict, optional
            Keyword arguments passed to the optimization function.
            Defaults {} but will set to
            {"method": "Powell", "x0"=x0} if optimizer is default.

        Returns
        -------
        res: scipy.optimize.OptimizeResult
            The result of the fitting (optimization)
        """
        self.model_empirical = model
        if x0 is None:
            str_sig_model = str(inspect.signature(model))
            nargs = len(str_sig_model.split(",")) - 1
            x0 = np.ones(nargs)

        if "method" not in optimizer_kwargs.keys():
            if "method" in inspect.signature(optimizer).parameters:
                optimizer_kwargs["method"] = "Powell"
        if "x0" not in optimizer_kwargs.keys():
            if "method" in inspect.signature(optimizer).parameters:
                optimizer_kwargs["x0"] = x0

        res = optimizer(
            kontrol.frequency_series.costs.cost_empirical_fit,
            args=(self.f, self.x, model, error_func, error_func_kwargs),
            **optimizer_kwargs)
        self.args_empirical_model = res.x
        self.x_empirical = model(self.f, *res.x)
        return res

    def fit_zpk(self, order, fit="x_empirical",
                padding=False, padding_order=1.,
                error_func=kontrol.core.math.log_mse,
                error_func_kwargs={},
                optimizer=scipy.optimize.differential_evolution,
                optimizer_kwargs={}):
        """Fit frequency series with a zpk model.

        Parameters
        ----------
        order: int
            The order of the ZPK model to be used.
        fit: str, optional
            Choose the data to be fit,
            choose from ["x", "x_empirical"].
        padding: boolean, optional
            Pad the data by continuous white noise.
            Defaults False.
        padding_order: float, optional
            The number of decades to be padded on both sides of the series.
            Defaults to 1.
        error_func: func(x1: array, x2: array) -> float
            The function that evaluate the error between arrays x1 and x2.
            Defaults to kontrol.core.math.log_mse, which evaluates the
            logarithmic mean square error.
        error_func_kwargs: dict, optional
            Keyword arguments passed to the error function.
            Use this to specify weighting function {"weight": weight_func}.
            Defaults to {}.
        optimizer: func, optional
            The optimization function which has the signature
            func(func, args, bounds, **kw) -> scipy.optimize.OptimizeResult.
            Defaults to scipy.optimize.differential_evolution.
        optimizer_kwargs: dict, optional
            Keyword arguments passed to the optimization function.
            Defaults {} but will set to
            {"workers": -1, "updating":"deferred"} if optimizer is default.

        Returns
        -------
        res: scipy.optimize.OptimizerResult
            The result of the optimization.

        Note
        ----
        For now, we support models with equal number of zeros and poles only.

        Best use with log-space frequency series or after using
        `fit_empirical()`.
        """
        if self._f_processed is None or self._x_processed is None:
            if fit == "x":
                f = self.f
                x = self.x
                if np.isclose(f[1]-f[0], f[2]-f[1]):
                    spacing = "linear"
                else:
                    spacing = "log"
            elif fit == "x_empirical":
                f = np.logspace(
                    np.log10(min(self.f)), np.log10(max(self.f)), len(self.f))
                x = self.model_empirical(f, *self.args_empirical_model)
                spacing = "log"
            else:
                raise ValueError("Choose fit from [\"x\" or \"x_empirical\"]")
            if padding:
                f, x = self._get_padded_data(f, x, padding_order, spacing)
            # Use these for zpk and tf fit.
            self._f_processed = f
            self._x_processed = x

        _optimizer_kwargs = dict(optimizer_kwargs)

        # Set defaults for the default optimizer
        if "workers" not in _optimizer_kwargs.keys():
            if "workers" in inspect.signature(optimizer).parameters:
                _optimizer_kwargs["workers"] = -1
        if "updating" not in _optimizer_kwargs.keys():
            if "updating" in inspect.signature(optimizer).parameters:
                _optimizer_kwargs["updating"] = "deferred"

        frequency_bounds = [(min(f), max(f))]*2*order  # FIXME
        gain_bound = [(min(x)*1e-1, max(x)*1e1)]    # FIXME
        bounds = frequency_bounds + gain_bound

        if "bounds" not in _optimizer_kwargs.keys():
            if "bounds" in inspect.signature(optimizer).parameters:
                _optimizer_kwargs["bounds"] = bounds

        res = optimizer(
            kontrol.frequency_series.costs.cost_zpk_fit,
            args=(f, x, error_func, error_func_kwargs),
            **_optimizer_kwargs)
        self.f_zpk = f
        self.x_zpk = kontrol.frequency_series.conversion.args2zpk(f=f, zpk_args=res.x)
        self.tf_zpk = kontrol.frequency_series.conversion.args2controltf(zpk_args=res.x)
        self.args_zpk_model = res.x
        return res

    def fit_tf(self, x0=None, log_var=True,
               error_func=kontrol.core.math.log_mse,
               error_func_kwargs={},
               optimizer=scipy.optimize.minimize,
               optimizer_kwargs={}):
        """Fit frequency series with a transfer function model

        Parameters
        ----------
        x0: array or None, optional
            The initial parameters defining the initial transfer function.
            A 1-D list of numerator and denominator coefficients,
            from higher order to lower order.
            Defaults None.
            If None, will use the result from the ZPK fitting.
        log_var: boolean, optional
            Optimize the log of variables instead.
            Useful for variables that have very large dynamic range
            Defaults True.
        error_func: func(x1: array, x2: array) -> float, optional
            The function that evaluate the error between arrays x1 and x2.
            Defaults to kontrol.core.math.log_mse, which evaluates the
            logarithmic mean square error.
        error_func_kwargs: dict, optional
            Keyword arguments passed to the error function.
            Use this to specify weighting function {"weight": weight_func}.
            Defaults to {}.
        optimizer: func, optional
            The optimization function which has the signature
            func(func, args, bounds, **kw) -> scipy.optimize.OptimizeResult.
            Defaults to scipy.optimize.minimize.
        optimizer_kwargs: dict, optional
            Keyword arguments passed to the optimization function.
            Defaults {} but will set to
            {"method": "Powell", "x0": x0} if optimizer is default.

        Returns
        -------
        res: scipy.optimize.OptimizerResult
            The result of the optimization.

        Note
        ----
        Only use this after using `fit_zpk()`.
        """
        if x0 is None:
            x0 = kontrol.frequency_series.conversion.tf2tf_args(tf=self.tf_zpk)
        if log_var:
            x0 = np.log(x0)

        _optimizer_kwargs = dict(optimizer_kwargs)
        if "method" not in _optimizer_kwargs.keys():
            if "method" in inspect.signature(optimizer).parameters:
                _optimizer_kwargs["method"] = "Powell"
        if "x0" not in _optimizer_kwargs.keys():
            if "method" in inspect.signature(optimizer).parameters:
                _optimizer_kwargs["x0"] = x0

        f = self._f_processed
        x = self._x_processed

        res = optimizer(
            kontrol.frequency_series.costs.cost_tf_fit,
            args=(f, x, error_func, error_func_kwargs, log_var),
            **_optimizer_kwargs)

        self.f_tf = f
        args_opt = res.x
        if log_var:
            args_opt = np.exp(args_opt)
        self.x_tf = kontrol.frequency_series.conversion.args2tf(f, args_opt)  # Frequency series
        self.tf = kontrol.frequency_series.conversion.tf_args2tf(args_opt)  # Transfer function
        self.args_tf_model = args_opt

        return res

    def _get_padded_data(self, f, x, padding_order, spacing="log"):
        """Calculate padded frequency axis.

        Parameters
        ----------
        f: array
            Frequency axis
        x: array
            Frequency series data.
        padding_order: float
            The number of decades to be padded on both sides of the series.
        spacing: str, optional
            The spacing of the specified data.
            Choose from ["log", "linear"].

        Returns
        -------
        f_pad: array
            Padded frequency axis.
        x_pad: array
            Padded frequency series data
        """
        len_new_f = ((np.log10(max(f)) + padding_order
                    - np.log10(min(f)) + padding_order)
                    / (np.log10(max(f)) - np.log10(min(f)))
                    * len(f))
        len_new_f = int(len_new_f)
        pad_width = int(((len_new_f)-len(f))/2)
        if spacing == "log":
            upper = np.log10(max(f)) + padding_order
            lower  = np.log10(min(f)) - padding_order
            f = np.logspace(
                np.log10(min(f)), np.log10(max(f)), len(f))
            log_f = np.log10(f)
            log_f_pad = np.pad(log_f, pad_width, "linear_ramp",
                end_values=(lower, upper))
            f_pad = 10**log_f_pad
        elif spacing == "linear":
            upper = max(f)*10**padding_order
            lower = min(f)/10**padding_order
            f_pad = np.pad(f, pad_width, "linear_ramp",
                end_values=(lower, upper))
        x_pad = np.pad(x, pad_width, "edge")
        return f_pad, x_pad



"""Utility function related to python-control library.
"""
import control
import numpy as np


def tfmatrix2tf(sys):
    """Convert a matrix of transfer functions to a MIMO transfer function.

    Parameters
    ----------
    sys: list of (list of control.xferfcn.TransferFunction)
        The transfer function matrix representing a MIMO system. sys[i][j]
        is the transfer function from the i+1 input port to the j+1 output
        port.
    Returns
    -------
    control.xferfcn.TransferFunction
        The transfer function of the MIMO system.
    """
    nums = list(np.zeros_like(sys))
    dens = list(np.zeros_like(sys))
    for i in range(len(sys)):
        for j in range(len(sys[i])):
            nums[i][j] = list(sys[i][j].num[0][0])
            dens[i][j] = list(sys[i][j].den[0][0])
        nums[i] = list(nums[i])
        dens[i] = list(dens[i])
    generalized_plant = control.tf(nums, dens)
    return generalized_plant


def zpk(zeros, poles, gain, unit='f', negate=True):
    """Zero-pole-gain definition of transfer function.

    Parameters
    ----------
    zeros: list of floats
        A list of the location of the zeros
    poles: list of floats
        A list of the location of the poles
    gain: float
        The static gain of the transfer function
    unit: str, optional
        The unit of the zeros, poles and the natural frequencies.
        Choose from ["f", "s", "Hz", "omega"].
        Defaults "f".
    negate: boolean, optional
        Negate zeros and poles in specification
        so negative sign is not needed for stable
        zeros and poles. Default to be True.

    Returns
    -------
    zpk_tf: control.xferfcn.TransferFunction
        The zpk defined transfer function

    Notes
    -----
    Refrain from specifying imaginary zeros and poles.
    Use kontrol.utils.sos() for second-order sections
    instead.
    The zero and poles are negated by default.
    """

    zeros = [float(z) for z in zeros]
    poles = [float(p) for p in poles]

    zeros = np.array(zeros)
    poles = np.array(poles)

    if unit in ["f", "Hz"]:
        for i in range(len(zeros)):
            zeros[i] = 2*np.pi*zeros[i]
        for i in range(len(poles)):
            poles[i] = 2*np.pi*poles[i]

    if negate is False:
        for i in range(len(zeros)):
            zeros[i] = -zeros[i]
        for i in range(len(poles)):
            poles[i] = -poles[i]

    zpk_tf = control.tf([gain],[1])
    for z in zeros:
        zpk_tf *= control.tf([1/z, 1], [1])
    for p in poles:
        zpk_tf *= control.tf([1], [1/p, 1])

    return zpk_tf


def sos(natural_frequencies=None, quality_factors=None, gain=1., unit="f"):
    r"""Second-order sections transfer fucntion.

    Parameters
    ----------
    natural_freuqnecies: array, optional
        List of natural frequencies.
        Defaults [].
    quality_factors: list, optional
        List of quality factors.
        Defaults [].
    gain: float, optional
        The gain of the transfer function.
        Defaults 1.
    unit: str, optional
        The unit of the zeros, poles and the natural frequencies.
        Choose from ["f", "s", "Hz", "omega"].
        Defaults "f".

    Returns
    -------
    sos: control.xferfcn.TransferFunction
        The product of all second-order sections.

    Note
    ----
    :math:`K\prod\left(s^2+\omega_n/Q\,s+\omega_n^2\right)`.
    """
    if natural_frequencies is None:
        natural_frequencies = []
    if quality_factors is None:
        quality_factors = []
    if len(natural_frequencies) != len(quality_factors):
        raise ValueError("Lenght of natural frequencies must match length of "
                         "quality factors.")

    natural_frequencies = np.array(natural_frequencies)
    quality_factors = np.array(quality_factors)
    if unit in ["f", "Hz"]:
        natural_frequencies = 2*np.pi*natural_frequencies

    s = control.tf("s")
    sos = control.tf([gain], [1])
    for wn, q in zip(natural_frequencies, quality_factors):
        sos *= (s**2 + wn/q*s + wn**2)/wn**2
    return sos


def convert_unstable_tf(control_tf):
    """Convert transfer function with unstable zeros and poles to tf without.

    Parameters
    ----------
    control_tf: control.xferfcn.TransferFunction
        The transfer function to be converted.

    Returns
    -------
    tf_new: control.xferfcn.TransferFunction
        control_tf with unstable zeros and poles flipped.

    Note
    ----
    Warning can occur when the order gets high, e.g. 100.
    This happens due to numerical accuracy,
    i.e. coefficients get astronomically high when order gets high.
    """
    zeros_unstable = control_tf.zeros()
    poles_unstable = control_tf.poles()
    zeros_stable = []
    poles_stable = []
    for zero in zeros_unstable:
        # Ignore zeros/poles that have negative imaginary part.
        if zero.imag != 0 and zero.imag < 0:
            pass
        elif zero.real > 0:
            zeros_stable.append(-zero.real + 1j*zero.imag)
        else:
            zeros_stable.append(zero.real + 1j*zero.imag)
    for pole in poles_unstable:
        # Ignore zeros/poles that have negative imaginary part.
        if pole.imag != 0 and pole.imag < 0:
            pass
        elif pole.real > 0:
            poles_stable.append(-pole.real + 1j*pole.imag)
        else:
            poles_stable.append(pole.real + 1j*pole.imag)

    tf_new = control.tf([1], [1])
    s = control.tf("s")
    for zero in zeros_stable:
        if zero.imag != 0:
            wn = np.sqrt(zero.real**2 + zero.imag**2)
            zeta = -zero.real/wn
            tf_new /= wn**2 / (s**2+2*zeta*wn*s+wn**2)
        else:
            tf_new *= s-zero.real
    for pole in poles_stable:
        if pole.imag != 0:
            wn = np.sqrt(pole.real**2 + pole.imag**2)
            zeta = -pole.real/wn
            tf_new *= wn**2 / (s**2+2*zeta*wn*s+wn**2)
        else:
            tf_new /= s-pole.real

    tf_new *= float(control_tf.dcgain())/float(tf_new.dcgain())

    return tf_new


def check_tf_equal(tf1, tf2, allclose_kwargs={}, minreal=True):
    """Check if two transfer functions are approximatedly equal by np.allclose.

    Parameters
    ----------
    tf1: control.xferfcn.TransferFunction
        Transfer function 1.
    tf2: control.xferfcn.TrasnferFunction
        Transfer function 2.
    allclose_kwargs: dict, optional
        Keyword arguments passed to np.allclose(), which is used to compare
        the list of poles and zeros and dcgain.
        Defaults {}.
    minreal: boolean
        Use control.minreal to remove canceling zeros and poles before
        comparison.

    Returns
    -------
    boolean
    """
    if minreal:
        tf1 = control.minreal(tf1, verbose=False)
        tf2 = control.minreal(tf2, verbose=False)
    zeros_close = np.allclose(tf1.zeros(), tf2.zeros(), **allclose_kwargs)
    poles_close = np.allclose(tf1.poles(), tf2.poles(), **allclose_kwargs)
    gain_close = np.allclose(tf1.dcgain(), tf2.dcgain(), **allclose_kwargs)
    return all([zeros_close, poles_close, gain_close])


def generic_tf(zeros=None, poles=None,
               zeros_wn=None, zeros_q=None,
               poles_wn=None, poles_q=None,
               dcgain=1., unit="f"):
    r"""Construct a generic transfer function object.

    Parameters
    ----------
    zeros: array, optional
        List of zeros in frequencies.
        Defaults [].
    poles: array, optional
        List of poles.
        Defaults [].
    zeros_wn: array, optional
        List of natural frequencies of numerator second-order sections.
    zeros_q: array, optional.
        List of Q-value of numerator second-order sections.
    poles_wn: array, optional
        List of natural frequencies of denominator second-order sections.
    poles_q: array, optional.
        List of Q-value of denominator second-order sections.
    dcgain: float, optional
        The DC gain of the transfer function.
        Defaults 1.
    unit: str, optional
        The unit of the zeros, poles and the natural frequencies.
        Choose from ["f", "s", "Hz", "omega"].
        Defaults "f".

    Returns
    -------
    tf: control.xferfcn.TransferFunction
        The transfer function.

    Note
    ----
    No negative sign is needed for negative zeros and poles.
    For instance, `generic_tf(poles=[1], unit="s")` means
    :math:`\frac{1}{s+1}`.
    """
    if zeros is None:
        zeros = []
    if poles is None:
        poles = []
    if zeros_wn is None:
        zeros_wn = []
    if zeros_q is None:
        zeros_q = []
    if poles_wn is None:
        poles_wn = []
    if poles_q is None:
        poles_q = []
    
    zeros = np.array(zeros)
    poles = np.array(poles)
    zeros_wn = np.array(zeros_wn)
    zeros_q = np.array(zeros_q)
    poles_wn = np.array(poles_wn)
    poles_q = np.array(poles_q)

    tf = zpk(zeros=zeros, poles=poles, gain=dcgain, unit=unit)
    tf *= sos(natural_frequencies=zeros_wn, quality_factors=zeros_q, unit=unit)
    tf /= sos(natural_frequencies=poles_wn, quality_factors=poles_q, unit=unit)
    return tf


def outliers(tf, f, unit="f"):
    """Returns a list of zeros and poles outside the frequency range.

    Parameters
    ----------
    tf: control.xferfcn.TransferFunction
        The transfer function.
    f: array
        The frequency axis of interest.
    unit: str, optional
        The unit of the zeros, poles and the natural frequencies.
        Choose from ["f", "s", "Hz", "omega"].
        Defaults "f".

    Returns
    -------
    outlier_zeros: array
        Zeros outside the frequency range.
    outlier_poles: array
        Poles outside the frequency range.

    Note
    ----
    We use the python-control package convention for
    the returned zeros and poles, so to preserve the type and format
    as much as possible for further processes.
    """
    outlier_zeros = []
    outlier_poles = []
    zeros = tf.zeros()
    poles = tf.poles()
    f = np.array(f)
    if unit in ["f", "Hz"]:
        f = f*2*np.pi
    f_min = min(f)
    f_max = max(f)

    for zero in zeros:
        fn = np.sqrt(zero.real**2 + zero.imag**2)
        if fn < f_min or fn > f_max:
            outlier_zeros.append(zero)
    for pole in poles:
        fn = np.sqrt(pole.real**2 + pole.imag**2)
        if fn < f_min or fn > f_max:
            outlier_poles.append(pole)
    outlier_zeros = np.array(outlier_zeros)
    outlier_poles = np.array(outlier_poles)
    return outlier_zeros, outlier_poles


def outlier_exists(tf, f, unit="f"):
    """Checks for zeros and poles outside the frequency range.

    Parameters
    ----------
    tf: control.xferfcn.TransferFunction
        The transfer function.
    f: array
        The frequency axis of interest.
    unit: str, optional
        The unit of the zeros, poles and the natural frequencies.
        Choose from ["f", "s", "Hz", "omega"].
        Defaults "f".

    Returns
    -------
    boolean
        If there's any zero or pole outside the frequency range.
    """
    outlier_zeros, outlier_poles = outliers(tf=tf, f=f, unit=unit)
    if len(outlier_zeros) + len(outlier_poles) > 0:
        return True
    else:
        return False


def tf_order_split(tf, max_order=20):
    """Split TransferFunction objects into multiple ones with fewer order.

    Parameters
    ----------
    tf : TransferFunction
        The transfer function
    max_order : int, optional
        The maxmium order of the split transfer functions.
        Defaults to 20 (the foton limit).

    Returns
    -------
    list of TransferFunction.
        The list of splitted transfer functions.
    """
    zeros = tf.zeros()
    poles = tf.poles()
    gain = tf.minreal().num[0][0][0]
    # for i in range(len(zeros)):
        # if zeros[i]
    nzero = len(zeros)
    npole = len(poles)
    niter_zero = int(np.ceil(nzero/max_order))
    niter_pole = int(np.ceil(npole/max_order))
    
    global order_running
    global zero_running
    global tf_zero_list
    global pole_running
    global tf_pole_list
    order_running = 0
    zero_running = []
    tf_zero_list = []
    pole_running = []
    tf_pole_list = []
    
    def put_tf_zeros_into_list():
        """Use zeros in zero_running to create an all zero transfer function
        and put store them into tf_zero_list, and reset order_running to 0.
        """
        s = control.tf("s")
        global order_running
        global zero_running
        global tf_zero_list
        simple_zeros = []
        zeros_wn = []
        zeros_q = []
        differentiator = 0
        gain = 1
        for z in zero_running:
            if z.imag == 0:
                if z.real == 0:
                    differentiator += 1
                    # gain *= 2*np.pi
                else:
                    gain *= -z.real
                    simple_zeros += [-z.real]
            else:
                wn = abs(z)
                q = wn / (-2*z.real)
                gain *= wn**2
                zeros_wn += [abs(z)]
                zeros_q += [q]
        tf = generic_tf(zeros=simple_zeros,
                        zeros_wn=zeros_wn, zeros_q=zeros_q,
                        dcgain=gain, unit="s")
        for i in range(differentiator):
            tf *= s
        tf_zero_list += [tf]
        order_running = 0
        zero_running = []

    def put_tf_poles_into_list():
        """Use poles in pole_running to create an all pole transfer function
        and put store them into tf_pole_list, and reset order_running to 0.
        """
        s = control.tf("s")
        global order_running
        global pole_running
        global tf_pole_list
        simple_poles = []
        poles_wn = []
        poles_q = []
        integrator = 0
        gain = 1
        for p in pole_running:
            if p.imag == 0:
                if p.real == 0:
                    integrator += 1
                    # gain /= 2*np.pi
                else:
                    gain /= -p.real
                    simple_poles += [-p.real]
            else:
                wn = abs(p)
                q = wn / (-2*p.real)
                gain /= wn**2
                poles_wn += [abs(p)]
                poles_q += [q]
        tf = generic_tf(poles=simple_poles,
                        poles_wn=poles_wn, poles_q=poles_q,
                        dcgain=gain, unit="s")
        for _ in range(integrator):
            tf *= 1/s
        tf_pole_list += [tf]
        order_running = 0
        pole_running = []

    # Put zeros into list until order reaches max_order.
    # And then make a transfer function out of the list,
    # and finally resets.
    for i in range(len(zeros)):
        if zeros[i].imag == 0:
            if order_running+1 > max_order:
                put_tf_zeros_into_list()
            order_running += 1
            zero_running += [zeros[i]]
        elif zeros[i].imag > 0:
            if order_running+2 > max_order:
                put_tf_zeros_into_list()
            order_running += 2
            zero_running += [zeros[i]]
        else:
            pass  ## Ignore negative part of the complex pair.
   
    ## Add the reminders
    put_tf_zeros_into_list()
            
    for i in range(len(poles)):
        if poles[i].imag == 0:
            if order_running+1 > max_order:
                put_tf_poles_into_list()
            order_running += 1
            pole_running += [poles[i]]
        elif poles[i].imag > 0:
            if order_running + 2 > max_order:
                put_tf_poles_into_list()
            order_running += 2
            pole_running += [poles[i]]
        else:
            pass  ## Ignore negative part of the complex pair.
    
    put_tf_poles_into_list()

    n_tf = max(len(tf_pole_list), len(tf_zero_list))
    tf_list = []
    for i in range(n_tf):
        # print(i)
        if i+1 > len(tf_pole_list):
            tf_list += [tf_zero_list[i]]
        elif i+1 > len(tf_zero_list):
            tf_list += [tf_pole_list[i]]
        else:
            tf_list += [tf_zero_list[i]*tf_pole_list[i]]
    tf_list[0] *= gain
    return tf_list


def clean_tf(tf, tol_order=5, small_number=1e-25):
    """Remove numerator/denominator coefficients that are small outliers

    Parameters
    ----------
    tf : TransferFunction
        The transfer function to be cleaned
    tol_order : float, optional
        If the coefficient is ``tol_order`` order smaller than
        the rest of the coefficients, then this coefficient is an outlier.
        Defaults 5.
    small_number : float, optional
        A small number to be added to the log10 in case 0 is encountered.

    Returns
    -------
    tf_cleaned : TransferFunction
        The cleaned transfer function.
    """
    num = tf.num[0][0].copy()
    den = tf.den[0][0].copy()
    log_num = np.log10(num+small_number)
    log_den = np.log10(den+small_number)
    num_mask = log_num.mean() - log_num > tol_order
    den_mask = log_den.mean() - log_den > tol_order
    num[num_mask] = 0
    den[den_mask] = 0
    tf_cleaned = control.tf(num, den)
    # tf_cleaned = tf_cleaned.minreal()
    return tf_cleaned


def clean_tf2(tf, tol_order=5, small_number=1e-25):
    """Remove zeros/poles that are outliers.

    Parameters
    ----------
    tf : TransferFunction
        The transfer function to be cleaned.
    tol_order : float, optional
        If the frequency of the zero/pole is ``tol_order'' order away from the
        mean order, then this zero/pole is an outlier.
        Defaults 5.
    small_number : float, optional
        A small number to be added to the log10() in case 0 is encountered

    Returns
    -------
    tf_cleaned : TransferFunction
        The cleaned transfer function.
    """
    zeros = tf.zeros().copy()
    poles = tf.poles().copy()
    wn_zeros = abs(zeros)
    wn_poles = abs(poles)
    log_wn_zeros = np.log10(wn_zeros+small_number)
    log_wn_poles = np.log10(wn_poles+small_number)
    mean_wn_zeros = np.mean(log_wn_zeros)
    mean_wn_poles = np.mean(log_wn_poles)
    mask_zeros = abs(log_wn_zeros-mean_wn_zeros) > tol_order
    mask_poles = abs(log_wn_poles-mean_wn_poles) > tol_order
    zeros[mask_zeros] = None
    poles[mask_poles] = None

    s = control.tf("s")
    tf_cleaned = control.tf([1], [1])
    for i, zero in enumerate(zeros):
        if mask_zeros[i]:  # Remove outliers
            continue
        if zero.imag < 0:  # Only treat positive imag part.
            continue
        elif zero.imag != 0 and zero.real != 0:
            wn, q = complex2wq(zero)
            tf_cleaned *= (s**2 + wn/q*s + wn**2) / wn**2
        elif zero.real == 0 and zero.imag != 0:
            wn = abs(zero)
            tf_cleaned *= (s**2 + wn**2) / wn**2
        elif zero.real != 0 and zero.imag == 0:
            tf_cleaned *= (s+abs(zero))/abs(zero)
        elif zero == 0:
            tf_cleaned *= s
        
    for i, pole in enumerate(poles):
        if mask_poles[i]:  # Remove outliers
            continue
        if pole.imag < 0:  # Only treat positive imag part.
            continue
        elif pole.imag != 0 and pole.real != 0:
            wn, q = complex2wq(pole)
            tf_cleaned /= (s**2 + wn/q*s + wn**2) / wn**2
        elif pole.real == 0 and pole.imag != 0:
            wn = abs(pole)
            tf_cleaned /= (s**2 + wn**2) / wn**2
        elif pole.real != 0 and pole.imag == 0:
            tf_cleaned /= (s+abs(pole))/abs(pole)
        elif pole == 0:
            tf_cleaned /= s

    # Match gain at 1 Hz
    gain_1hz = abs(tf(1j*2*np.pi*1))
    tf_cleaned *= gain_1hz / abs(tf_cleaned(1j*2*np.pi*1))
    # tf_cleaned = tf_cleaned.minreal()
    return tf_cleaned


def clean_tf3(tf, tol_order=5, small_number=1e-25):
    """Remove first coefficient if it is much smaller than the second
    
    Parameters
    ----------
    tf : TransferFunction
        Transfer function to be cleaned
    tol_order : float, optional
        First coefficient is removed if it is ``tol_order'' order
        smaller than the second coefficient.
        Remove only if there exists such coefficient in both
        numerator and denominator.
    small_number : float, optional
        A small number to be added to the log10() in case 0 is encountered.
    
    Returns
    -------
    tf_cleaned : TransferFunction
        The cleaned TransferFunction
    """
    num = tf.num[0][0].copy()
    den = tf.den[0][0].copy()
    if (abs(np.log10(num[0]) - np.log10(num[1])) > tol_order
        and abs(np.log10(den[0]) - np.log10(den[1])) > tol_order):
        #
        tf_cleaned = control.tf(tf.num[0][0][1:], tf.den[0][0][1:])
    else:
        tf_cleaned = tf
    return tf_cleaned


def complex2wq(complex_frequency):
    """Convert a complex frequency to frequency and Q-factor.

    Parameters
    ----------
    complex_frequency : complex
        Complex frequency in rad/s.

    Returns
    -------
    wn : float
        The frequency in rad/s
    q : float
        The Q factor.
    """
    wn = abs(complex_frequency)
    q = -wn / (2*complex_frequency.real)
    q = abs(q)
    return wn, q



"""Model base class for curve fitting.
"""


class Model:
    """Model base class for curve fitting"""
    def __init__(self, args=None, nargs=None, log_args=False):
        """Constructor

        Parameters
        ----------
        args : array or None, optional
            The model parameters.
            Defaults to None.
        nargs : int, optional
            The number of model parameters needed to define the model.
            Defaults to None.
        log_args : boolean, optional
            If true,
            model parameters passed to the model are assumed to be passed
            through a log10() function.
            So, when the real parameters will be assumed to be
            10**args instead.
            Defaults to False.
        """
        self.log_args = log_args
        self._args = None
        self._nargs = None
        self.args = args
        self.nargs = nargs

    def __call__(self, x, args=None, **kwargs):
        """Use self._x2y(x, **kwargs) to give y values from x.

        This is a low level function that shouldn't be used by users.

        Parameters
        ----------
        x : array
            Independent variables
        args : array or None, optional
            Model parameters.
            Defaults to None.
            If not specified, use self.args.
            If specified, set self.args to args.
        **kwargs
            Keyword arguments passed to self._x2y().

        Returns
        -------
        y : array
            f(x) = y values.
        """
        if args is None:
            args = self.args
        else:
            self.args = args
        if self.args is None:
            raise ValueError("Please specify model parameters by the "
                             "arg argument or by setting self.arg.")
        return self._x2y(x, **kwargs)

   # def __add__(self, model2):
   #      """Summing the model by another model, returns a CompositeModel.
        
   #      Parameters
   #      ----------
   #      model2 : kontrol.curvefit.model.Model
   #          The other model to be summed.

   #      Returns
   #      -------
   #      kontrol.curvefit.model.CompositeModel
   #          The sum of two models.
   #      """
   #      if self.nargs is None or model2.nargs None:
   #          raise ValueError("The number of arguments must be defined "
   #                           "before the models can be combined.")
   #      split_index = self.nargs
   #      return CompositeModel(
   #          model1=self, model2=model2, split_index=split_index, operator="+")

    def _x2y(self, x, **kwargs):
        """Convert independent variables to dependent variables.

        This should be defined independent in child classes.
        This by default returns ``x`` itself, i.e. f(x) = x.

        Parameters
        ----------
        x : array
            The independent variables.
        **kwargs
            Keyword arguments.

        Returns
        -------
        y : array
            f(x) = x.
        """
        y = x
        return y

    @property
    def args(self):
        """Model parameters"""
        return self._args

    @args.setter
    def args(self, _args):
        """args.setter"""
        if self.log_args and _args is not None:
            _args = 10**_args
        self._args = _args
        # self.nargs = len(_args)

    @property
    def nargs(self):
        """Number of model parameters"""
        return self._nargs
    
    @nargs.setter
    def nargs(self, _nargs):
        """nargs.setter"""
        self._nargs = _nargs
        



"""Dynamic mode decomposition class"""
import numpy as np
import scipy


class DMD:
    """Dynamic mode decomposition class

    Parameters
    ----------
    snapshot_1 : array
        The 2-D snapshot array.
    snapshot_2 : array, optional
        The 2-D snapshot array at next time step.
        If not provided, ``snapshot_1[:, :-1]`` becomes snapshot_1 and
        ``snapshot_2[:, 1:]`` becomes ``snapshot_2``.
        Defaults None.
    truncation_value : float, optional
        The truncation value (order/rank) of the system.
        Defaults None.
    dt : float, optional
        The time difference between two snapshots.
        Defaults 1.
    run : bool, optional
        Run dynamic mode decomposition upon construction.
        Computes DMD modes, Reduced-order model, etc.
        truncation_value must be specified for this to work.
        Defaults False.

    Attributes
    ----------
    snapshot_1 : array
        The 2-D snapshot matrix
    snapshot_2 : array
        The 2-D snapshot matrix at next time step.
    truncation_value : int
        The truncation value (order/rank) of the system.
    dt : float
        The time difference between two snapshots.
    u : array
        The left singular vector of the snapshot_1 matrix.
    vh : array
        The right singular vector (conjugate-transposed) of the snapshot_1
        matrix.
    sigma : array
        The singular values of the snapshot_1 matrix
    u_truncated : array
        The truncated left singular vector of the snapshot_1 matrix.
    vh_truncated : array
        The truncated right singular vector (conjugate-transposed) of the
        snapshot_1 matrix.
    sigma_truncated : array
        The truncated singular values of the snapshot_1 matrix
    A_reduced : array
        The reduced-order model of the dynamics.
    w_reduced : array
        The eigenvalues of the reduced-order model.
    v_reduced : array
        The eigenvectors of the reduced-order model.
    dmd_modes : array
        The DMD modes.
    complex_frequencies : array
        The complex frequencies of the modes.
    v_constant : array
        The constant vector of the time dynamics, determined by
        initial conditions.
    time_dynamics : array
        The time dynamics of the ODE solution.
    prediction : array
        The predicted time series.
    """
    def __init__(self, snapshot_1, snapshot_2=None,
                 truncation_value=None, dt=1, run=False):
        """Constructor

        Parameters
        ----------
        snapshot_1 : array
            The 2-D snapshot matrix.
        snapshot_2 : array, optional
            The 2-D snapshot matrix at next time step.
            If not provided, ``snapshot_1[:, :-1]`` becomes snapshot_1 and
            ``snapshot_2[:, 1:]`` becomes ``snapshot_2``.
            Defaults None.
        truncation_value : float, optional
            The truncation value (order/rank) of the system.
            Defaults None.
        dt : float, optional
            The time difference between two snapshots.
            Defaults 1.
        run : bool, optional
            Run dynamic mode decomposition upon construction.
            Computes DMD modes, Reduced-order model, etc.
            truncation_value must be specified for this to work.
            Defaults False.
        """
        # Initializing attributes
        self._snapshot_1 = None
        self._snapshot_2 = None
        self._u = None
        self._vh = None
        self._sigma = None
        self._u_truncated = None
        self._vh_truncated = None
        self._sigma_truncated = None
        self._A_reduced = None
        self._w_reduced = None
        self._v_reduced = None
        self._dmd_modes = None
        self._complex_frequencies = None
        self._v_constant = None
        self._time_dynamics = None
        self._dt = None
        self._complex_frequencies = None
        self._prediction = None
        self._truncation_value = None

        self.truncation_value = truncation_value
        self.dt = dt

        if snapshot_2 is None:
            self.snapshot_2 = snapshot_1[:, 1:]
            self.snapshot_1 = snapshot_1[:, :-1]
        else:
            self.snapshot_1 = snapshot_1
            self.snapshot_2 = snapshot_2

        if run:
            self.run()

    def run(self):
        """Run the DMD algorithm, compute dmd modes and complex frequencies"""
        self.svd()
        self.low_rank_approximation()
        self.compute_reduced_model()
        self.eig_reduced_model()
        self.compute_dmd_modes()
        self.compute_complex_frequencies()

    @property
    def snapshot_1(self):
        """Snapshot 1"""
        return self._snapshot_1

    @snapshot_1.setter
    def snapshot_1(self, _snapshot_1):
        """snapshot_1.setter"""
        self._snapshot_1 = _snapshot_1
        
    @property
    def snapshot_2(self):
        """Snapshot 1"""
        return self._snapshot_2

    @snapshot_2.setter
    def snapshot_2(self, _snapshot_2):
        """snapshot_2.setter"""
        self._snapshot_2 = _snapshot_2
        
    def svd(self):
        """Decompose the snapshot 1
        
        Returns
        -------
        u : array
            Unitary matrix having left singular vectors as columns.
        sigma : array
            The singular values.
        vh : array
            Unitary matrix having right singular vectors as rows.
        """
        u, sigma, vh = scipy.linalg.svd(self.snapshot_1)
        self.u = u
        self.vh = vh
        self.sigma = sigma
        return u, sigma, vh

    @property
    def u(self):
        """Left singular vectors"""
        return self._u

    @u.setter
    def u(self, _u):
        """u.setter"""
        self._u = _u

    @property
    def vh(self):
        """Right singular vectors"""
        return self._vh

    @vh.setter
    def vh(self, _vh):
        """vh.setter"""
        self._vh = _vh

    @property
    def sigma(self):
        """Singular values"""
        return self._sigma

    @sigma.setter
    def sigma(self, _sigma):
        """sigma.setter"""
        self._sigma = _sigma

    def low_rank_approximation(self, truncation_value=None):
        """Truncate the svd.
        
        Paramters
        ---------
        truncation_value : int, optional
            The truncation value (order/rank) of the system.
            Specify as self.truncation_value or via the constuctor option.
            Defaults None.
        
        Returns
        -------
        u_truncated : array
            Truncated unitary matrix having left singular vectors as columns.
        sigma_truncated : array
            Truncated singular values.
        vh_truncated : array
            Truncated unitary matrix having right singular vectors as rows.
        """
        if truncation_value is None:
            if self.truncation_value is None:
                raise ValueError("Please specify truncation_value.")
            truncation_value = self.truncation_value
        else:
            self.truncation_value = truncation_value
        #TODO Add option to automatically select truncation value

        if self.u is None or self.sigma is None or self.vh is None:
            self.svd()

        u_truncated = self.u[:, :truncation_value]
        sigma_truncated = self.sigma[:truncation_value]  # This is a 1-D array.
        sigma_truncated = scipy.linalg.diagsvd(
            sigma_truncated, len(sigma_truncated), len(sigma_truncated))
        vh_truncated = self.vh[:truncation_value, :]
        self.u_truncated = u_truncated
        self.sigma_truncated = sigma_truncated  # This is a square matrix
        self.vh_truncated = vh_truncated
        return u_truncated, sigma_truncated, vh_truncated
    
    @property
    def truncation_value(self):
        """Truncation value (order/rank) of the system"""
        return self._truncation_value

    @truncation_value.setter
    def truncation_value(self, _truncation_value):
        """truncation_value.setter"""
        self._truncation_value = _truncation_value

    @property
    def u_truncated(self):
        """Left singular vectors (truncated)"""
        return self._u_truncated

    @u_truncated.setter
    def u_truncated(self, _u_truncated):
        """u_truncated.setter"""
        self._u_truncated = _u_truncated

    @property
    def vh_truncated(self):
        """Right singular vectors (truncated)"""
        return self._vh_truncated

    @vh_truncated.setter
    def vh_truncated(self, _vh_truncated):
        """vh_truncated.setter"""
        self._vh_truncated = _vh_truncated

    @property
    def sigma_truncated(self):
        """Singular values truncated"""
        return self._sigma_truncated

    @sigma_truncated.setter
    def sigma_truncated(self, _sigma_truncated):
        """sigma_truncated.setter"""
        self._sigma_truncated = _sigma_truncated

    def compute_reduced_model(self):
        """Compute the reduced-order model
        
        Returns
        -------
        A_reduced : array
            Matrix representing the reduced-order model.
        """
        if (self.u_truncated is None or self.vh_truncated is None
                or self.sigma_truncated is None):
            self.low_rank_approximation()

        u_h = self.u_truncated.T.conjugate()
        v = self.vh_truncated.T.conjugate()
        sigma_inv = scipy.linalg.inv(self.sigma_truncated)
        A_reduced = u_h @ self.snapshot_2 @ v @ sigma_inv
        self.A_reduced = A_reduced
        self.eig_reduced_model()  # Compute eigendecomposition as well.
        return A_reduced

    @property
    def A_reduced(self):
        """Reduced-order model"""
        return self._A_reduced

    @A_reduced.setter
    def A_reduced(self, _A_reduced):
        """A_reduced.setter"""
        self._A_reduced = _A_reduced
    
    def eig_reduced_model(self):
        """Eigen decomposition of the reduced-order model

        Returns
        -------
        w_reduced : array
            The list of eigenvalues.
        v_reduced : array
            Eigenvalues as columns.
        """
        if self.A_reduced is None:
            self.compute_reduced_model()

        w_reduced, v_reduced = scipy.linalg.eig(self.A_reduced)
        self.w_reduced = w_reduced
        self.v_reduced = v_reduced
        return w_reduced, v_reduced

    def compute_dmd_modes(self):
        """Compute the DMD modes
        
        Returns
        -------
        dmd_modes : array
            The DMD modes.
        """
        if self.vh_truncated is None or self.sigma_truncated is None:
            self.low_rank_approximation()
        if self.v_reduced is None:
            self.compute_reduced_model()

        v_truncated = self.vh_truncated.T.conjugate()
        sigma_inv = scipy.linalg.inv(self.sigma_truncated)
        # Exact DMD
        dmd_modes = self.snapshot_2 @ v_truncated @ sigma_inv @ self.v_reduced
        # Standard DMD
        # dmd_modes = self.u_truncated @ self.v_reduced
        self.dmd_modes = dmd_modes
        return dmd_modes

    def compute_complex_frequencies(self, dt=None):
        """Compute the complex frequencies

        Parameters
        ----------
        dt : float, optional
            The time spacing between snapshots.
            Specified as self.dt or in constructor option.
            Defaults 1.

        Returns
        -------
        complex_frequencies : array
            Array of complex frequencies
        """
        if dt is None and self.dt is None:
            self.dt = 1
        elif self.dt is None:
            self.dt = dt
        if self.w_reduced is None:
            self.eig_reduced_model()
            
        complex_frequencies = np.log(self.w_reduced) / self.dt
        self.complex_frequencies = complex_frequencies
        return complex_frequencies

    @property
    def dt(self):
        """Time difference between the snapshots"""
        return self._dt

    @dt.setter
    def dt(self, _dt):
        """dt.setter"""
        self._dt = _dt

    @property
    def complex_frequencies(self):
        """Complex frequencies"""
        return self._complex_frequencies

    @complex_frequencies.setter
    def complex_frequencies(self, _complex_frequencies):
        """complex_frequencies.setter"""
        self._complex_frequencies = _complex_frequencies

    def predict(self, t, t_constant=None, i_constant=0):
        """Predict the future states given a time array
        
        Parameters
        ----------
        t : array
            The time array
        t_constant : float, optional
            Time at which the constant vector is calculated.
            Defaults to be ``t[0]`` if not given.
        i_constant : int, optional
            Index of the ``snapshot_1`` at which the
            constant vector is calculated.
            Defaults 0.

        Returns
        -------
        prediction : array
            Predicted states.

        Notes
        -----
        The constant vector is the vector that evolves and is
        evaluated using the initial/boundary values.
        Set ``t_constant`` and ``i_constant`` to where the
        prediction of the time series begins.
        """
        # Make complex frequencies into a diagonal matrix.
        diag_complex_frequencies = np.diag(self.complex_frequencies)
        
        # Expand the array into another dimension as time series.
        complex_frequencies_series = np.repeat(
            diag_complex_frequencies[:, :, np.newaxis], len(t), axis=2)
        
        # Compute the exponentials.
        exponentials = np.exp(complex_frequencies_series * t)

        # Clear the non-diagonal elementes
        # I need to understand how the following diagonalization code works.
        diag = np.einsum("iik->ik", exponentials)
        save = diag.copy()
        exponentials[...] = 0
        diag[...] = save
        # exponentials is diagonalized.
        
        # Compute the constant vector.
        # Assume t=0 at snapshot[:, 0].
        # v_constant, _, _, _ = np.linalg.lstsq(
        #     self.dmd_modes, self.snapshot_1[:, 0], rcond=None)

        if t_constant is None:
            t_constant = t[0]
        v_constant, _, _, _ = np.linalg.lstsq(
            self.dmd_modes\
                @ np.diag(np.exp(self.complex_frequencies*t_constant)),
            self.snapshot_1[:, i_constant], rcond=None)
        self.v_constant = v_constant
        
        # Compute time dynamics exp(omega*t) @ b
        time_dynamics = np.einsum("ijk,j->ik", exponentials, v_constant)
        self.time_dynamics = time_dynamics

        # Prediction = dmd_modes @ exp(omega*t) @ b
        prediction = self.dmd_modes @ self.time_dynamics
        self.prediction = prediction

        return prediction

    @property
    def v_constant(self):
        """Constant vector of the ODE solution"""
        return self._v_constant

    @v_constant.setter
    def v_constant(self, _v_constant):
        """v_constant.setter"""
        self._v_constant = _v_constant

    @property
    def time_dynamics(self):
        """Time dynamics of the ODE solution"""
        return self._time_dynamics

    @time_dynamics.setter
    def time_dynamics(self, _time_dynamics):
        """time_dynamics.setter"""
        self._time_dynamics = _time_dynamics

    @property
    def prediction(self):
        """The ODE solution"""
        return self._prediction

    @prediction.setter
    def prediction(self, _prediction):
        """prediction.setter"""
        self._prediction = _prediction




"""Complementary filter class for sythesis
"""
import control
import numpy as np


import kontrol.core.math
import kontrol.core.controlutils
import kontrol.complementary_filter.synthesis
import kontrol.transfer_function


class ComplementaryFilter():
    r"""Complementary filter synthesis class.

    Parameters
    ----------
    noise1 : TransferFunction
        Sensor noise 1 transfer function model.
        A transfer function that has magnitude response matching the
        amplitude spectral density of noise 1.
    noise2 : TransferFunction
        Sensor noise 2 transfer function model.
        A transfer function that has magnitude response matching the
        amplitude spectral density of noise 2.
    weight1 : TransferFunction, optional
        Weighting function 1.
        Frequency dependent specification for noise 1.
        Defaults None.
    weight2 : TransferFunction, optional
        Weighting function 2.
        Frequency dependent specification for noise 2.
        Defaults None.
    filter1 : TransferFunction, optional
        The complementary filter for noise 1.
        Defaults None.
    filter2 : TransferFunction, optional
        The complementary filter for noise 2.
        Defaults None.
    f : array, optional
        The frequency axis in Hz for evaluating the super sensor noise.
        Defaults None.

    Methods
    -------
    h2synthesis()
        Complementary filter synthesis that minimizes the
        :math:`\mathcal{H}_2`-norm
        of the super sensor noise. The noise must be specified
        before using this method.
    hinfsynthesis()
        Complementary filter synthesis that minimizes the
        :math:`\mathcal{H}_\infty`-norm
        of the super sensor noise. The noise must be specified
        before using this method.
    
    Notes
    -----
    This is a utility class for complementary filter synthesis,
    sensor noise estimation and analysis.
    To use synthesis methods, specify ``noise1`` and ``noise2`` as
    the transfer function modesl for the sensor noises, and
    specify ``weight1`` and ``weight2`` as the frequency dependent
    specification for the two sensor noises.

    References
    ----------
    .. [1]
        T. T. L. Tsang, T. G. F. Li, T. Dehaeze, C. Collette.
        Optimal Sensor Fusion Method for Active Vibration Isolation Systems in
        Ground-Based Gravitational-Wave Detectors.
        https://arxiv.org/pdf/2111.14355.pdf
    """
    def __init__(self, noise1=None, noise2=None, weight1=None, weight2=None,
                 filter1=None, filter2=None, f=None):
        """Constructor

        Parameters
        ----------
        noise1 : TransferFunction
            Sensor noise 1 transfer function model.
            A transfer function that has magnitude response matching the
            amplitude spectral density of noise 1.
        noise2 : TransferFunction
            Sensor noise 2 transfer function model.
            A transfer function that has magnitude response matching the
            amplitude spectral density of noise 2.
        weight1 : TransferFunction, optional
            Weighting function 1.
            Frequency dependent specification for noise 1.
            Defaults None.
        weight2 : TransferFunction, optional
            Weighting function 2.
            Frequency dependent specification for noise 2.
            Defaults None.
        filter1 : TransferFunction, optional
            The complementary filter for noise 1.
            Defaults None.
        filter2 : TransferFunction, optional
            The complementary filter for noise 2.
            Defaults None.
        f : array, optional
            The frequency axis in Hz for evaluating the super sensor noise.
            Defaults None.
        """
        self._noise1 = None
        self._noise2 = None
        self._weight1 = None
        self._weight2 = None
        self._filter1 = None
        self._filter2 = None
        self._f = None
        self.noise1 = noise1
        self.noise2 = noise2
        self.weight1 = weight1
        self.weight2 = weight2
        self.filter1 = filter1
        self.filter2 = filter2
        self._f = f

    def h2synthesis(self, clean_filter=True, **kwargs):
        """Synthesize complementary filters using H2 synthesis.
        
        Returns
        -------
        filter1 : TransferFunction
            The complementary filter filtering noise 1.
        filter2 : TransferFunction
            The complementary filter filtering noise 2.
        clean_filter : boolean, optional
            Remove small outlier coefficients from filters.
            Defaults True.
        **kwargs
            Keyword arguments passed to kontrol.TransferFunction.clean()
        """
        func = kontrol.complementary_filter.synthesis.h2complementary
        return self._synthesis(func=func, clean_filter=clean_filter, **kwargs)

    def hinfsynthesis(self, clean_filter=True, **kwargs):
        """Synthesize complementary filters using H-inifinity synthesis.
        
        Returns
        -------
        filter1 : TransferFunction
            The complementary filter filtering noise 1.
        filter2 : TransferFunction
            The complementary filter filtering noise 2.
        clean_filter : boolean, optional
            Remove small outlier coefficients from filters.
            Defaults True.
        **kwargs
            Keyword arguments passed to kontrol.TransferFunction.clean()
        """
        func = kontrol.complementary_filter.synthesis.hinfcomplementary
        return self._synthesis(func=func, clean_filter=clean_filter, **kwargs)

    def _synthesis(self, func, clean_filter=True, **kwargs):
        """Generic complementary filter synthesis function.

        Synthesize the complementary filter using the function and
        re-initialize kontrol.transfer_function.TransferFunction
        with a 2-input-1-output control.xferfcn.TransferFunction.

        Parameters
        ----------
        func: function
            The function that takes two noise models and returns
            a pair of complementary filters.
            It should have a siguration of
            func(TransferFunction, TransferFunction) ->
            (TransferFunction, TransferFunction).
        clean_filter : boolean, optional
            Remove small outlier coefficients from filters.
            Defaults True.
        **kwargs
            Keyword arguments passed to kontrol.TransferFunction.clean()
        """
        filter1, filter2 = func(
            self.noise1, self.noise2, self.weight1, self.weight2, **kwargs)
        self.filter1 = filter1
        self.filter2 = filter2
        if clean_filter:
            self.filter1.clean(**kwargs)
            self.filter2 = 1 - self.filter1
        return self.filter1, self.filter2

    @property
    def noise1(self):
        """Transfer function model of sensor noise 1.
        """
        return self._noise1

    @noise1.setter
    def noise1(self, _noise1):
        """noise1 setter

        Parameters
        ----------
        noise1 : TransferFunction
        """
        self._noise1 = _noise1

    @property
    def noise2(self):
        """Transfer function model of sensor noise 2.
        """
        return self._noise2

    @noise2.setter
    def noise2(self, _noise2):
        """noise2 setter

        Parameters
        ----------
        noise2 : TransferFunction
        """
        self._noise2 = _noise2

    def noise_super(self, f=None, noise1=None, noise2=None,
                    filter1=None, filter2=None):
        """Compute and return predicted the ASD of the super sensor noise

        Parameter
        ---------
        f : array, optional
            The frequency axis in Hz.
            Use self.f if None.
            Defaults None.
        noise1 : array, optional
            The amplitude spectral density of noise 1.
            Use self.noise1 and self.f to estimate if None.
            Defaults None.
        noise2 : array, optional
            The amplitude spectral density of noise 2.
            Use self.noise1 and self.f to estimate if None.
            Defaults None.
        filter1 : TransferFunction, optional
            The complementary filter for filtering noise1
            Use self.filter1 if not specified.
            Defaults None
        filter2 : TransferFunction, optional
            The complementary filter for filtering noise1
            Use self.filter2 if not specified.
            Defaults None
        
        Returns
        -------
        array
            The amplitude spectral density of the super sensor noise.
        """
        if f is None:
            if self.f is None:
                raise ValueError("self.f is not specified.")
            f = self.f
        if filter1 is None:
            if self.filter1 is None:
                raise ValueError("self.filter1 is not specified. "
                                 "Please specify self.filter1 or synthesize.")
            filter1 = self.filter1
        if filter2 is None:
            if self.filter2 is None:
                raise ValueError("self.filter2 is not specified. "
                                 "Please specify self.filter2 or synthesize.")
            filter2 = self.filter2
        if noise1 is None:
            if self.noise1 is None:
                raise ValueError("noise1 is not specified. "
                                 "Please specify noise1 or self.noise1.")
            noise1 = abs(self.noise1(1j*2*np.pi*f))
        if noise2 is None:
            if self.noise2 is None:
                raise ValueError("noise2 is not specified. "
                                 "Please specify noise2 or self.noise2.")
            noise2 = abs(self.noise2(1j*2*np.pi*f))
        noise1_filtered = abs(filter1(1j*2*np.pi*f)) * noise1
        noise2_filtered = abs(filter2(1j*2*np.pi*f)) * noise2
        return kontrol.core.math.quad_sum(noise1_filtered, noise2_filtered)

    @property
    def filter1(self):
        """First complementary filter.
        """
        return self._filter1

    @filter1.setter
    def filter1(self, _filter1):
        """filter1 setter.

        Parameters
        ----------
        _filter1 : TransferFunction
        """
        if _filter1 is None:
            self._filter1 = _filter1
        else:
            self._filter1 = kontrol.transfer_function.TransferFunction(
                _filter1)

    @property
    def filter2(self):
        """Second complementary filter.
        """
        return self._filter2

    @filter2.setter
    def filter2(self, _filter2):
        """filter2 setter.

        Parameters
        ----------
        _filter2 : TransferFunction
        """
        if _filter2 is None:
            self._filter2 = _filter2
        else:
            self._filter2 = kontrol.transfer_function.TransferFunction(
                _filter2)

    @property
    def f(self):
        """The frequency axis in Hz.
        """
        return self._f

    @f.setter
    def f(self, _f):
        """Frequency axis setter.

        Parameters
        ----------
        _f: array
            The frequency axis.
        """
        self._f = _f



"""Cost functions for fitting
"""
import numpy as np

import kontrol.core.math
import kontrol.frequency_series.conversion



def cost_empirical_fit(args, f, x, model,
                       error_func=kontrol.core.math.log_mse,
                       error_func_kwargs={}):
    """Cost function for frequency series empirical fitting.

    Parameters
    ----------
    args: array
        Empirical model parameters.
    f: array
        Frequency axis.
    x: array
        Frequency series data.
    model: func(f, a, b, c, ...) -> array
        The model function whose first argument is the frequency axis and
        the rest being an arbitrary number of model parameters to be fit.
    error_func: func(x1: array, x2: array) -> float
        The function that evaluate the error between arrays x1 and x2.
        Defaults to kontrol.core.math.log_mse, which evaluates the
        logarithmic mean square error.
    error_func_kwargs: dict
        Keyword arguments passed to the error function.

    Returns
    -------
    cost: float
        The cost.
    """
    x_model = model(f, *args)
    cost = error_func(x, x_model, **error_func_kwargs)
    return cost


def cost_zpk_fit(zpk_args, f, x,
                 error_func=kontrol.core.math.log_mse,
                 error_func_kwargs={}):
    """The cost function for fitting a frequency series with zero-pole-gain.

    Parameters
    ----------
    zpk_args: array
        A 1-D list of zeros, poles, and gain.
        Zeros and poles are in unit of Hz.
    f: array
        The frequency axis.
    x: array
        The frequecy series data.
    error_func: func(x1: array, x2: array) -> float
        The function that evaluate the error between arrays x1 and x2.
        Defaults to kontrol.core.math.log_mse, which evaluates the
        logarithmic mean square error.
    error_func_kwargs: dict, optional
        Keyword arguments passed to the error function.
        Defaults {}.

    Returns
    -------
    cost: float
        The cost.
    """
    x_zpk = abs(
        kontrol.frequency_series.conversion.args2zpk(f=f, zpk_args=zpk_args))
    cost = error_func(x, x_zpk, **error_func_kwargs)
    return cost


def cost_tf_fit(tf_args, f, x,
                error_func=kontrol.core.math.log_mse,
                error_func_kwargs={},
                log_var=True):
    """The cost funcion for gitting a frequency series with transfer function.

    Parameters
    ----------
    tf_args: array
        A 1-D list of numerator and denominator coefficients,
        from higher order to lower order.
    f: array
        The frequency axis.
    x: array
        The frequecy series data.
    error_func: func(x1: array, x2: array) -> float, optional
        The function that evaluate the error between arrays x1 and x2.
        Defaults to kontrol.core.math.log_mse, which evaluates the
        logarithmic mean square error.
    error_func_kwargs: dict, optional
        Keyword arguments passed to the error function.
        Defaults {}.
    log_var: boolean, optional
        Optimize the log of variables instead.
        Useful for variables that have very large dynamic range
        Defaults True.

    Returns
    -------
    cost: float
        The cost.
    """
    if log_var:
        tf_args = np.exp(tf_args)
    x_tf = abs(
        kontrol.frequency_series.conversion.args2tf(f=f, tf_args=tf_args))
    cost = error_func(x, x_tf, **error_func_kwargs)
    return cost



"""Spectral analysis related function library.
"""
import numpy as np
import scipy.signal


def two_channel_correlation(psd, coh):
    r"""Noise estimation from two identical sensors' readouts.

    Parameters
    ----------
    psd : array
        The power spectral density of the readout of the sensor
    coh : array
        The coherence between readout of the sensor and another identical
        sensor.

    Returns
    -------
    noise : array
        Power spectral density of the estimated noise.

    Notes
    -----
    The PSD of the noise is computed as

    .. math::
        P_{nn}(f) = P_{x_1x_1}(f)\left(1-C_{x_1x_2}(f)^{\frac{1}{2}}\right)\,,

    where :math:`P_{x_1x_1}(f)` is the power spectral density of the readout
    and :math:`C_{x_1x_2}(f)` is the coherence
    between the two sensor readouts.

    References
    ----------
    .. [1]
        Aaron Barzilai, Tom VanZandt, and Tom Kenny.
        Technique for measurement of the noise of a sensor in the
        presence of large background signals. Review of Scientific Instruments,
        69:2767–2772, 07 1998.
    """
    noise = psd * (1 - coh**0.5)
    return noise


def three_channel_correlation(psd1, psd2=None, psd3=None,
                              csd12=None, csd13=None,
                              csd21=None, csd23=None,
                              csd31=None, csd32=None,
                              returnall=True):
    r"""Noise estimation from three sensors' readouts.

    Parameters
    ----------
    psd1 : array
        The power spectral density of the readout of the first sensor.
    psd2 : array, optional
        The power spectral density of the readout of the second sensor.
        Defaults ``None``.
    psd3 : array, optional
        The power spectral density of the readout of the third sensor.
        Defaults ``None``.
    csd12 : array, optional
        Cross power spectral density between readout 1 and 2.
        If not specified, this will be estimated as ``psd1*psd2/csd21``.
        Default ``None``.
    csd13 : array, optional
        Cross power spectral density between readout 1 and 3.
        If not specified, this will be estimated as ``psd1*psd3/csd31``.
        Default ``None``.
    csd21 : array, optional
        Cross power spectral density between readout 2 and 1.
        If not specified, this will be estimated as ``psd2*psd1/csd12``.
        Default ``None``.
    csd23 : array, optional
        Cross power spectral density between readout 2 and 3.
        If not specified, this will be estimated as ``psd2*psd3/csd32``.
        Default ``None``.
    csd31 : array, optional
        Cross power spectral density between readout 3 and 1.
        If not specified, this will be estimated as ``psd1*psd3/csd13``.
        Default ``None``.
    csd32 : array, optional
        Cross power spectral density between readout 3 and 1.
        If not specified, this will be estimated as ``psd3*psd2/csd23``.
        Default ``None``.
    returnall : boolean, optional
        If ``True``, return all three noise estimations.
        If ``False``, return noise estimation of first sensor only.
        Defaults True.

    Returns
    -------
    noise1 : array
        Power spectral density of the estimated noise in `psd1`.
    noise2 : array, optional
        Power spectral density of the estimated noise in `psd2`.
        Returns only if ``returnall==True``
    noise3 : array, optional
        Power spectral density of the estimated noise in `psd3`.
        Returns only if ``returnall==True``

    Notes
    -----
    The PSD of the noise in `psd1` is then computed as

    .. math::
       P_{n_1n_1}(f) = \left\lvert P_{x_1x_1}(f) -
       \frac{P_{x_1x_3}(f)}{P_{x_2x_3}(f)}P_{x_2x_1}\right\vert\,,

    If ``returnall`` is ``True``, at least ``psd1``, ``psd2``, ``psd3``,
    (``csd13`` or ``csd31``), (``csd23`` or ``csd32``), and
    (``csd12`` and ``csd21``) must be provided.

    References
    ----------
    .. [2]
        R. Sleeman, A. Wettum, and J. Trampert.
        Three-channel correlation analysis: A new technique to measure
        instrumental noise of digitizers and seismic sensors.
        Bulletin of the Seismological Society of America, 96:258–271, 2006.
    """
    if csd12 is None:
        if all([csd21 is not None, psd1 is not None, psd2 is not None]):
            csd12 = psd1*psd2/csd21
    if csd13 is None:
        if all([csd31 is not None, psd1 is not None, psd3 is not None]):
            csd13 = psd1*psd3/csd31
    if csd21 is None:
        if all([csd12 is not None, psd2 is not None, psd1 is not None]):
            csd21 = psd2*psd1/csd12
    if csd23 is None:
        if all([csd32 is not None, psd2 is not None, psd3 is not None]):
            csd23 = psd2*psd3/csd32
    if csd31 is None:
        if all([csd13 is not None, psd3 is not None, psd1 is not None]):
            csd31 = psd3*psd1/csd13
    if csd32 is None:
        if all([csd23 is not None, psd3 is not None, psd2 is not None]):
            csd32 = psd3*psd2/csd23
    if returnall:
        if any([psd2 is None, psd3 is None,
                csd12 is None, csd13 is None,
                csd21 is None, csd23 is None,
                csd31 is None, csd32 is None]):
            raise ValueError("If returnall is True,"
                             " at least psd2, psd3"
                             " (csd13 or csd31),"
                             " (csd23 or csd32), and"
                             " (csd12 and csd21) must be provided.")
        noise1 = three_channel_correlation(
            psd1, csd13=csd13, csd23=csd23, csd21=csd21, returnall=False)
        noise2 = three_channel_correlation(
            psd2, csd13=csd21, csd23=csd31, csd21=csd32, returnall=False)
        noise3 = three_channel_correlation(
            psd3, csd13=csd32, csd23=csd12, csd21=csd13, returnall=False)
        return noise1, noise2, noise3
    else:
        if any([csd13 is None, csd23 is None, csd21 is None]):
            raise ValueError("(csd13 or csd31), (csd23 or csd32),"
                             " and (csd21 or csd12) must be provided")
        return abs(psd1-csd13/csd23*csd21)


def asd2ts(asd, f=None, fs=None, t=None, window=None, zero_mean=True):
    """Simulate time series from amplitude spectral density

    Parameters
    ----------
    asd : array
        The amplitude spectral density.
    f : array, optional
        The frequency axis of the amplitude spectral density.
        This is used to calculate the sampling frequency.
        If ``fs`` is not specified, ``f`` must be specified.
        Defaults ``None``.
    fs : float, optional
        The sampling frequency in Hz.
        If ``f`` is specified, the last element of ``f`` will
        be treated as the Nyquist frequency so the double of
        it would be the sampling frequency.
        Defaults ``None``.
    t : array, optional
        The desired time axis for the time series.
        If ``t`` is specified, then the time series will
        be interpolated using ``numpy.interp()`` assuming
        that the time series is periodic.
        Default ``None``.
    window : array, optional
        The FFT window used to compute the amplitude spectral density.
        Defaults ``None``.
    zero_mean : boolean, optional
        Returns a time series with zero mean.
        Defaults ``True``.

    Returns
    -------
    t : array
        Time axis.
    time_series : array
        The time series

    Notes
    -----
    Using a custom time axis is not recommended as interpolation leads
    to distortion of the signal.
    To extend the signal in time, it's recommended to repeat the
    original time series instead.
    """
    if f is not None:
        fs = f[-1]*2  # Double the Nyquist frequency.
    elif fs is None:
        raise ValueError("Either f or fs must be specified.")

    if window is None:
        window = np.ones_like(asd)

    s2 = np.sum(window**2)  # One of the normalization factor.
    # Compute absolute value of the FFT
    abs_ym = asd*np.sqrt(fs*s2)
    # Assume ASD is noise and generate random phase for the FFT
    random_phase = np.random.random(len(abs_ym))*360-180
    # Convert abs(FFT) to FFT by adding random phase
    ym = abs_ym * np.exp(1j*random_phase)
    ts = np.fft.irfft(ym)

    if zero_mean:
        ts -= np.mean(ts)

    t_original = np.linspace(0, (len(ts)-1)/fs, len(ts))
    if t is None:
        t = t_original
    else:
        period = t_original[-1] - t_original[0]
        ts = np.interp(t, xp=t_original, fp=ts, period=period)
    return t, ts


def asd2rms(asd, f=None, df=1., return_series=True):
    r"""Calculate root-mean-squared value from amplitude spectral density

    Parameters
    ----------
    asd : array
        The amplitude spectral density
    f : array, optional
        The frequency axis.
        Defaults ``None``.
    df : float, optional
        The frequency spacing.
        Defaults 1.
    return_series : bool, optional
        Returns the RMS as a frequency series,
        where each value is the integrated RMS from highest frequency.
        If False, returns a single RMS value for the whole spectrum.
        Defaults True.

    Returns
    -------
    array
        The integrated RMS series, if ``return_series==True``.
    float
        The integrated RMS value, if ``return_series==False``.
    
    Notes
    -----
    The integrated RMS series is defined as

    .. math::
    
       x_\mathrm{RMS}(f) = \int_\infty^{f}\,x(f')^2\,df'\,,

    where :math:`x(f)` is the amplitude spectral density.

    When ``return_series`` is False, only :math:`x_\mathrm{RMS}(0)` is
    returned.
    """
    # Flip the ASD to integrate from high frequency to low.
    # if f is not None:
    #     f_inv = -np.flip(f)  # Minus sign necessary as integration limits
    #                          # are flipped
    # else:
    #     f_inv = None
    # asd_inv = np.flip(asd)
    if not return_series:
        rms = np.sqrt(np.trapz(y=asd**2, x=f, dx=df))
    else:
        rms = np.zeros_like(asd)
        for i in range(len(rms)):
            if f is None:
                rms[i] = np.sqrt(np.trapz(y=asd[i:]**2, dx=df))
            else:
                rms[i] = np.sqrt(np.trapz(y=asd[i:]**2, x=f[i:]))
        # rms = np.flip(rms)
    return rms


def pad_below_minima(series, pad_index=0, **kwargs):
    """Pad series below local minima

    Parameters
    ----------
    series : array
        The series to be padded.
    pad_index : int, optional
        The index of the local minima.
        Defaults to 0.
    **kwargs
        Keyword arguments passed to scipy.signal.argrelmin()

    Returns
    -------
    series_padded : array
        The padded series
    """
    i_min_list = scipy.signal.argrelmin(series, **kwargs)
    i_min = i_min_list[0][pad_index]
    series_padded = series.copy()
    series_padded[:i_min+1] = series[i_min]
    return series_padded


def pad_above_minima(series, pad_index=-1, **kwargs):
    """Pad series above local minima

    Parameters
    ----------
    series : array
        The series to be padded.
    pad_index : int, optional
        The index of the local minima.
        Defaults to -1.
    **kwargs
        Keyword arguments passed to scipy.signal.argrelmin()

    Returns
    -------
    series_padded : array
        The padded series
    """
    i_min_list = scipy.signal.argrelmin(series, **kwargs)
    i_min = i_min_list[0][pad_index]
    series_padded = series.copy()
    series_padded[i_min:] = series[i_min]
    return series_padded


def pad_below_maxima(series, pad_index=0, **kwargs):
    """Pad series below local maxima

    Parameters
    ----------
    series : array
        The series to be padded.
    pad_index : int, optional
        The index of the local maxima.
        Defaults to 0.
    **kwargs
        Keyword arguments passed to scipy.signal.argrelmin()

    Returns
    -------
    series_padded : array
        The padded series
    """
    i_max_list = scipy.signal.argrelmax(series, **kwargs)
    i_max = i_max_list[0][pad_index]
    series_padded = series.copy()
    series_padded[:i_max+1] = series[i_max]
    return series_padded


def pad_above_maxima(series, pad_index=-1, **kwargs):
    """Pad series above local maxima

    Parameters
    ----------
    series : array
        The series to be padded.
    pad_index : int, optional
        The index of the local maxima.
        Defaults to -1.
    **kwargs
        Keyword arguments passed to scipy.signal.argrelmin()

    Returns
    -------
    series_padded : array
        The padded series
    """
    i_max_list = scipy.signal.argrelmax(series, **kwargs)
    i_max = i_max_list[0][pad_index]
    series_padded = series.copy()
    series_padded[i_max:] = series[i_max]
    return series_padded




"""Common math library
"""
import numpy as np


def mse(x1, x2, weight=None):
    """Mean square error

    Parameters
    ----------
    x1: array
        Array
    x2: array
        Array
    weight: array or None, optional
        weighting function.
        If None, defaults to np.ones_like(x1)

    Returns
    -------
    float
        Mean square error between arrays x1 and x2.
    """
    if weight is None:
        weight = np.ones_like(x1)
    return np.mean(np.square(np.subtract(x2, x1)*weight))


def log_mse(x1, x2, weight=None, small_multiplier=1e-6):
    """Logarithmic mean square error/Mean square log error

    Parameters
    ----------
    x1: array
        Array
    x2: array
        Array
    weight: array or None, optional
        weighting function.
        If None, defaults to `np.ones_like(x1)`
    small_multiplier: float, optional
        `x1` will be modified by `x1`+`small_multiplier`*`np.min(x2)` if
        0 exists in `x1`.
        Same goes to `x2`.
        If 0 both exists in `x1` and `x2`, raise.

    Returns
    -------
    float
        Logarithmic mean square error between arrays x1 and x2.
    """
    if 0 in x1 and 0 in x2:
        raise ValueError("zero value exists in both 'x1' and 'x2'"
                         "Can't take np.log10()")
    elif 0 in x1:
        x1 += small_multiplier * np.min(x2)
    elif 0 in x2:
        x2 += small_multiplier * np.min(x1)
    return mse(x1=np.log10(x1), x2=np.log10(x2), weight=weight)


def quad_sum(*spectra):
    """Takes any number of same length spectrum and returns the quadrature sum.

    Parameters
    ----------
        *spectra:
            Variable length argument list of the spectra.

    Returns
    -------
        qs: np.ndarray
            The quadrature sum of the spectra.

    """
    qs = np.zeros_like(spectra[0])
#     print(args[0])
    for i in spectra:
        for j in range(len(i)):
            qs[j] = np.sqrt(qs[j]**2 + i[j]**2)
    return qs


def polyval(p, x):
    y = np.zeros(x.shape, dtype=x.dtype)
    for i, v in enumerate(p):
        y *= x
        y += v
    return y



"""Cost function base class
"""
class Cost:
    """Cost function base class.
    """
    def __init__(self, error_func, error_func_kwargs=None):
        """Constructor

        Parameters
        ----------
        error_func : func(array, array, **error_func_kwargs) -> float.
            The error function.
            It takes two arrays and evaluate the error.
        error_func_kwargs : dict or None, optional
            Keyword arguments the will be passed to ``error_func``
            Defaults to None.
        """
        self._error_func = None
        self._error_func_kwargs = None
        self.error_func = error_func
        self.error_func_kwargs = error_func_kwargs

    def __call__(self, args, model, xdata, ydata, model_kwargs=None):
        """Evaluate the cost function.

        Parameters
        ----------
        args : array
            Arguments passed to the model
        model : func(xdata: array, args: array, **model_kwargs) -> array
            Model used to fit the data.
        x_data : array
            Independent variable data.
        y_data : array
            Dependent variable data.
        model_kwargs : dict or None, optional
            Keyword arguments passed to the ``model``.

        Returns
        -------
        float
            Evaluated error function.
        """
        if model_kwargs is None:
            model_kwargs = {}
        y_model = model(xdata, args, **model_kwargs)
        error_func = self.error_func
        error_func_kwargs = self.error_func_kwargs
        error = error_func(ydata, y_model, **error_func_kwargs)
        return error

    @property
    def error_func(self):
        """The error function."""
        return self._error_func

    @error_func.setter
    def error_func(self, _error_func):
        """error_func setter"""
        self._error_func = _error_func

    @property
    def error_func_kwargs(self):
        """Keyword arguments passed to error_func"""
        return self._error_func_kwargs

    @error_func_kwargs.setter
    def error_func_kwargs(self, _error_func_kwargs):
        """error_func_kwargs setter"""
        if _error_func_kwargs is None:
            self._error_func_kwargs = {}
        else:
            self._error_func_kwargs = _error_func_kwargs



"""Predefined regulator library. """
import control
import numpy as np

import kontrol


def proportional_integral_derivative(kp=0, ki=0, kd=0):
    r"""PID control build on proportional_derivative().
    
    Parameters
    ----------
    kp : float, optional
        The proportional control gain.
        Defaults to 0.
    ki : float, optional
        The integral control gain.
        Defaults to 0.
    kd : float, optional
        Defaults to 0.

    Returns
    -------
    TransferFunction
        The PID controller.
        
    Notes
    -----
    The PID controller is defined as

    .. math::
       
       K_\mathrm{PID}(s) = K_p + K_i/s + K_d s\,.
    """
    s = control.tf("s")
    return kontrol.TransferFunction(kp + ki/s + kd*s)


def pid(kp=0, ki=0, kd=0):
    """Alias of proportional_integral_derivative()"""
    return proportional_integral_derivative(kp, ki, kd)


def low_pass(cutoff, order=1, **kwargs):
    r"""Simple low-pass filter

    Parameters
    ----------
    cutoff : float
        Cutoff frequency (Hz)
    order : int, optional
        The order of the filter.
        Defaults to be 1.
    **kwargs
        Keyword arguments holder. Not passed to anywhere.
    
    Returns
    -------
    TransferFunction
        The low-pass filter.

    Notes
    -----
    The low-pass filter is defined as

    .. math::
       
       L(s) = \left(\frac{2\pi f_c}{s+2\pi f_c}\right)^n\,,

    where :math:`f_c` is the cutoff frequency (Hz), :math:`n` is the order
    of the filter.
    """
    s = control.tf("s")
    lp = ((2*np.pi*cutoff) / (s + 2*np.pi*cutoff))**order
    return kontrol.TransferFunction(lp)


def notch(frequency, q, depth=None, depth_db=None, **kwargs):
    r"""Notch filter defined in Foton.

    Parameters
    ----------
    frequency : float
        The notch frequency (Hz).
    q : float
        The quality factor.
    depth : float, optional
        The depth of the notch filter (magnitude).
        If not specified, ``depth_db`` will be used.
        Defaults None.
    depth_db : float, optional
        The depth of the notch filter (decibel).
        If not specified, ``depth`` will be used instead.
        Defaults None.

    Returns
    -------
    TransferFunction
        The notch filter

    Notes
    -----
    The notch filter is defined by Foton, as
    
    .. math::
       
       N(s) = \frac{s^2 + (2\pi f_n)/(dQ/2)s + (2\pi f_n)^2}
       {s^2 + (2\pi f_n)/(Q/2)s + (2\pi f_n)^2}\,,

    where :math:`f_n` is the notch frequency, :math:`q` is the quality factor
    , and :math:`d` is the depth.
    """
    if depth is None and depth_db is None:
        raise ValueError("Either depth or depth_db must be specified")
    if depth is None:
        depth = 10**(depth_db/20)

    s = control.tf("s")
    wn = 2*np.pi*frequency
    qp = q/2
    qz = qp*depth
    n = (s**2 + wn/qz*s + wn**2) / (s**2 + wn/qp*s + wn**2)
    return kontrol.TransferFunction(n) 



"""Conversion functions for frequency series fitting.

This is a lower level module which users are not expect to use.
"""
import control
import numpy as np


def args2zpk(f, zpk_args):
    """Returns an array of ZPK defined transfer function evaluted at frequency f.

    Parameters
    ----------
    f: array
        The frequency axis.
    zpk_args: array
        A 1-D list of zeros, poles, and gain.
        Zeros and poles are in unit of Hz.

    Returns
    -------
    zpk: array
        A complex array.
    """
    s = 1j*2*np.pi*f
    zeros = zpk_args[:int(len(zpk_args)/2)]
    poles = zpk_args[int(len(zpk_args)/2):len(zpk_args)-1]
    gain = zpk_args[-1]
    zpk = np.prod([(s/(2*np.pi*z)+1)/(s/(2*np.pi*p)+1)
                  for z,p in zip(zeros, poles)], axis=0)
    zpk *= gain
    return zpk


def args2controltf(zpk_args):
    """Convert a list of zeros, poles, and gain to control.tf

    Parameters
    ----------
    zpk_args: array
        A 1-D list of zeros, poles, and gain.
        Zeros and poles are in unit of Hz.
    """
    s = control.tf("s")
    zeros = zpk_args[:int(len(zpk_args)/2)]
    poles = zpk_args[int(len(zpk_args)/2):len(zpk_args)-1]
    gain = zpk_args[-1]
    zpk = np.prod([(s/(2*np.pi*z)+1)/(s/(2*np.pi*p)+1)
                  for z,p in zip(zeros, poles)], axis=0)
    zpk *= gain
    return zpk


def args2tf(f, tf_args):
    """Returns an array of transfer function evaluted at frequency f.

    Parameters
    ----------
    f: array
        The frequency axis.
    tf_args: array
        A 1-D list of numerator and denominator coefficients,
        from higher order to lower order.

    Returns
    -------
    array
        A complex array.
    """
    s = 1j*2*np.pi*f
    num_coef = tf_args[:int(len(tf_args)/2)]
    den_coef = tf_args[int(len(tf_args)/2):len(tf_args)]
#     print(num_coef)
#     print(den_coef)
    num = np.sum([np.flip(num_coef)[i] * s**i for i in range(len(num_coef))], axis=0)
    den = np.sum([np.flip(den_coef)[i] * s**i for i in range(len(den_coef))], axis=0)
    return num/den


def tf2tf_args(tf):
    """Returns an array of tf_args used for fitting.

    Parameters
    ----------
    tf: control.xferfcn.TransferFunction
        The transfer function.

    Returns
    -------
    tf_args: array
        A 1-D list of numerator and denominator coefficients,
        from higher order to lower order.
    """
    num = tf.num[0][0]
    den = tf.den[0][0]
    tf_args = np.concatenate((num, den))
    return tf_args


def tf_args2tf(tf_args):
    """Returns a transfer function from tf_args.

    Parameters
    ----------
    tf_args: array
        A 1-D list of numerator and denominator coefficients,
        from higher order to lower order.

    Returns
    -------
    tf: control.xferfcn.TransferFunction
        The transfer function.
    """
    num_coef = tf_args[:int(len(tf_args)/2)]
    den_coef = tf_args[int(len(tf_args)/2):len(tf_args)]
    tf = control.tf(num_coef, den_coef)
    return tf
